'''Module to handle KVQ files (Sega Smash Pack, Sega Puzzle Pack)'''

def extract(kvq_bytes, encode_string):
    '''Extract a sega pack kvq'''
    scramble = 6
    out_file = bytearray()
    for index, encoded_byte in enumerate(kvq_bytes[8:]):
        encode_string_character = encode_string[index % len(encode_string)]
        decoded_byte = ((encoded_byte ^ encode_string_character ^ 0x80) - scramble) & 0xFF
        out_file.append(decoded_byte)
        scramble += 3
    return out_file



'''Implementation of breakers: Breakers Collection'''
import logging
import os
from gex.lib.tasks import helpers
from gex.lib.tasks.basetask import BaseTask

logger = logging.getLogger('gextoolbox')

class DisneyClassicsTask(BaseTask):
    '''Implements breakers: Breakers Collection'''
    _task_name = "breakers"
    _title = "Breakers Collection"
    _details_markdown = ''''''
    _default_input_folder = helpers.gen_steam_app_default_folder(
        "Breakers Collection")
    _input_folder_desc = "Breakers Collection Steam folder"

    def get_out_file_info(self):
        '''Return a list of output files'''
        return {
            "files": self._metadata['out']['files'],
            "notes": self._metadata['out']['notes']
        }

    def execute(self, in_dir, out_dir):
        in_files = self.read_all_datafiles(in_dir)

        output_files = []
        output_files.append(self._handle_breakers(in_files))
        output_files.append(self._handle_breakrev(in_files))
        
        for out_file_entry in output_files:
            _ = self.verify_out_file(out_file_entry['filename'], out_file_entry['contents'])
            out_path = os.path.join(out_dir, out_file_entry['filename'])
            with open(out_path, "wb") as out_file:
                out_file.write(out_file_entry['contents'])
                
        logger.info("Processing complete.")

    def _handle_breakers(self, in_files):
        out_files = {}
        out_files['230-p1.p1'] = in_files['230-m68k.swbin']['contents']
        out_files['230-c1.c1'] = in_files['230-p1.spr']['contents']
        out_files['230-c2.c2'] = in_files['230-p2.spr']['contents']
        out_files['230-c3.c3'] = in_files['230-p3.spr']['contents']
        out_files['230-c4.c4'] = in_files['230-p4.spr']['contents']
        out_files['230-s1.s1'] = in_files['230-sfix.dat']['contents']
        out_files['230-v1.v1'] = in_files['230-snd1.pcm']['contents']
        out_files['230-v2.v2'] = in_files['230-snd2.pcm']['contents']
        out_files['230-m1.m1'] = in_files['230-z80.bin']['contents']
        return {'filename': 'breakers.zip', 'contents': helpers.build_zip(out_files)}

    def _handle_breakrev(self, in_files):
        out_files = {}
        out_files['245-p1.p1'] = in_files['245-m68k.swbin']['contents']
        out_files['245-c1.c1'] = in_files['245-p1.spr']['contents']
        out_files['245-c2.c2'] = in_files['245-p2.spr']['contents']
        out_files['245-c3.c3'] = in_files['245-p3.spr']['contents']
        out_files['245-c4.c4'] = in_files['245-p4.spr']['contents']
        out_files['245-c5.c5'] = in_files['245-p5.spr']['contents']
        out_files['245-c6.c6'] = in_files['245-p6.spr']['contents']
        out_files['245-s1.s1'] = in_files['245-sfix.dat']['contents']
        out_files['245-v1.v1'] = in_files['245-snd1.pcm']['contents']
        out_files['245-v2.v2'] = in_files['245-snd2.pcm']['contents']
        out_files['245-m1.m1'] = in_files['245-z80.bin']['contents']
        return {'filename': 'breakrev.zip', 'contents': helpers.build_zip(out_files)}



'''Implementation of basic zip splice task (for arcade ROMs, primarily)'''
import logging
import os
from gex.lib.tasks.basetask import BaseTask
from gex.lib.tasks import helpers
from gex.lib.utils.blob import transforms

logger = logging.getLogger('gextoolbox')

class ZipSpliceTask(BaseTask):
    '''Implements basic zip splice task (for arcade ROMs, primarily)'''

    def get_out_file_info(self):
        '''Return a list of output files'''
        return {
            "files": self._metadata['out']['files'],
            "notes": self._metadata['out']['notes']
        }

    def execute(self, in_dir, out_dir):
        '''Splice out the ROM files'''
        resolved_file = self.read_datafile(in_dir, self._metadata['in']['files']['source'])
        source_data = resolved_file['contents']

        extractable_roms = [x for x in self._metadata['out']['files'] if x['status'] != 'no-rom']
        for file_meta in extractable_roms:
            logger.info(f"Extracting {file_meta['game']}...")
            zip_files = {}

            for filename, section in file_meta['zip_files'].items():
                if section.get('type') == 'placeholder':
                    length = int(section['length'], 16)
                    file_content = bytearray(b'\x00' * length)
                else:
                    start = int(section['start'], 16)
                    length = int(section['length'], 16)
                    file_content = transforms.cut(source_data, start, length=length)

                patches = section.get('patch')
                if patches and len(patches.keys()) > 0:
                    for loc, data in patches.items():
                        file_content[int(loc)] = data.to_bytes(1, 'little')[0]

                zip_files[filename] = file_content

            game_data = helpers.build_zip(zip_files)
            filename = file_meta['filename']
            _ = self.verify_out_file(filename, game_data)
            out_path = os.path.join(out_dir, filename)
            with open(out_path, "wb") as out_file:
                out_file.write(game_data)

        logger.info("Processing complete.")



from os.path import dirname, basename, isfile, split
import glob

# add simple modules
modules = glob.glob(dirname(__file__)+"/*.py")
__all__ = [ basename(f)[:-3] for f in modules if isfile(f) and not basename(f).startswith('__')] # exclude __init__.py

# add complex modules
modules = glob.glob(dirname(__file__)+"/*/__init__.py")
__all__ += [ split(dirname(f))[1] for f in modules ]

__all__.sort()



'''Implmentation of saga: Collection of SaGa Final Fantasy Legend'''
import logging
import os
import UnityPy
from gex.lib.tasks.basetask import BaseTask
from gex.lib.tasks import helpers

logger = logging.getLogger('gextoolbox')

class SagaTask(BaseTask):
    '''Implments saga: Collection of SaGa Final Fantasy Legend'''
    _task_name = "saga"
    _title = "Collection of SaGa Final Fantasy Legend"
    _details_markdown = '''
These are extracted from the Unity asset bundle files.
See https://github.com/farmerbb/RED-Project/issues/39 for more info.
'''
    _out_file_notes = {}
    _default_input_folder = helpers.gen_steam_app_default_folder("Sa・Ga COLLECTION")
    _input_folder_desc = "Collection of SaGa Steam folder"

    def get_out_file_info(self):
        '''Return a list of output files'''
        return {
            "files": self._metadata['out']['files'],
            "notes": self._metadata['out']['notes']
        }

    def execute(self, in_dir, out_dir):
        # for each output file entry
        for out_file_entry in self._metadata['out']['files']:
            pkg_name = out_file_entry['extract']['in_file']
            # Check the status of it
            if out_file_entry['status'] == 'no-rom':
                logger.info(f"Skipping {pkg_name} - cannot extract...")
            else:
                logger.info(f"Extracting {pkg_name}...")

                # read the matching input file
                in_file_entry = self._metadata['in']['files'][pkg_name]
                loaded_file = self.read_datafile(in_dir, in_file_entry)

                # load the archive
                unity_bundle = UnityPy.load(loaded_file['contents'])

                # Get the rom asset entry
                rom_asset = unity_bundle.container.get(out_file_entry['extract']['archive_path'])
                rom_data = rom_asset.read().script

                if rom_asset is None:
                    logger.warning("Could not find rom asset in archive.")
                else:
                    _ = self.verify_out_file(out_file_entry['filename'], rom_data)

                    with open(os.path.join(out_dir, out_file_entry['filename']), "wb") as out_file:
                        out_file.write(rom_data)
        logger.info("Processing complete.")




'''Common operations for blobs (binary contents) and arrays of blobs'''
import itertools
from typing import Callable
from bitarray import bitarray


def cut(contents: bytes, start: int, length: int=None, end: int=None) -> bytearray:
    '''Select a chunk out of a blob'''
    if length is None and end is None:
        raise Exception(
            "Splice out needs a length or end value, but received neither.")
    elif length is not None and end is not None:
        raise Exception(
            "Splice out needs a length or end value, but received both.")
    elif end is None:
        end = start + length

    return bytearray(contents[start:end])

def merge(chunks: list[bytes]) -> bytearray:
    '''Merge an array of blobs into a single blob'''
    new_content = bytearray()
    for chunk in chunks:
        new_content += chunk
    return new_content

def pad(contents: bytes, new_length: int, pad_byte=b'\0') -> bytearray:
    '''Pad a blob to a specific size'''
    new_content = bytearray()
    new_content.extend(contents)
    pad_size = new_length - len(contents)
    new_content.extend(bytearray(pad_byte * pad_size))
    return new_content

def custom_split(contents: bytes, chunk_sizes: list[int]) -> list[bytearray]:
    '''Split a blob into an array of blobs with specified sizes'''
    if len(contents) != sum(chunk_sizes):
        raise Exception(
            f'Custom split chunk_sizes {sum(chunk_sizes)} != input len {len(contents)}')

    num_chunks = len(chunk_sizes)
    start_offset = 0
    temp_chunks = []
    for i in range(0, num_chunks):
        temp_chunks.append(
            bytearray(contents[start_offset:start_offset+chunk_sizes[i]]))
        start_offset = start_offset + chunk_sizes[i]
    return temp_chunks


def equal_split(contents: bytes, num_chunks: int=None, chunk_size: int=None) -> list[bytearray]:
    '''Split a blob in an array of blobs with equal sizes'''
    if num_chunks is None and chunk_size is None:
        raise Exception("Equal split num_chunks or chunk_Size required")

    if num_chunks is not None and chunk_size is not None:
        if num_chunks * chunk_size != len(contents):
            raise Exception(
                "Equal split num_chunks * chunk_size != content length.")
    elif num_chunks is None:
        num_chunks = len(contents)//chunk_size
    else:
        chunk_size = len(contents)//num_chunks

    start_offset = 0
    temp_chunks = []
    for _ in range(0, num_chunks):
        temp_chunks.append(
            bytearray(contents[start_offset:start_offset+chunk_size]))
        start_offset = start_offset + chunk_size
    return temp_chunks


def interleave(chunks: list[bytes], word_size: int) -> bytearray:
    '''Interleave an array of blobs together into a single blob, word_size bytes at a time'''
    if len(chunks) < 2:
        raise Exception("Interleave requires at least 2 chunks to interleave.")

    chunk_length = len(chunks[0])
    for chunk in chunks:
        if len(chunk) != chunk_length:
            raise Exception("Interleave requires chunks of the same size.")

    combined_length = len(chunks) * chunk_length

    interleave_group_length = len(chunks) * word_size
    num_interleave_groups = combined_length//interleave_group_length
    new_contents = bytearray()
    for i in range(0, num_interleave_groups):
        offset = i * word_size
        for chunk in chunks:
            new_contents += chunk[offset:offset+word_size]
    return new_contents


def deinterleave(contents: bytes, num_ways: int, word_size: int) -> list[bytearray]:
    '''Deinterleave a blob into an array of num_ways blobs, word_size bytes at a time'''
    interleave_group_length = num_ways * word_size
    temp_chunks = []
    for j in range(0, num_ways):
        # Make a compress flag array of the combined group length.
        # This 1s only for the appropriate bytes for this chunk.
        flag_arr = [0] * interleave_group_length
        flag_arr[j*word_size:(j+1)*word_size] = [1 for val in flag_arr[j *
                                                                       word_size:(j+1)*word_size]]

        # Use compress() to select bytes based on the flag array repeated via cycle()
        temp_chunks.append(bytearray(itertools.compress(
            contents, itertools.cycle(flag_arr))))
    return temp_chunks

def deinterleave_nibble(contents: bytes, num_ways: int, bit_prefix: bitarray=bitarray('0000')) -> list[bytearray]:
    '''Deinterleave a blob into an array of num_ways blobs, 4 bits at a time; must be an even num_ways'''
    interleave_group_length = int(num_ways * .5)
    num_interleave_groups = len(contents) // interleave_group_length
    temp_bit_chunks = []
    for j in range(0, num_ways):
        temp_bit_chunks.append(bitarray())
    curr_bits = bitarray()
    for i in range(0, num_interleave_groups):
        curr_bytes = contents[i*interleave_group_length:(i+1)*interleave_group_length]
        curr_bits.clear()
        curr_bits.frombytes(curr_bytes)

        for j in range(0, num_ways):
            temp_bit_chunks[j].extend(bit_prefix + curr_bits[j*4:(j+1)*4])

    return [bytearray(temp_bit_chunk.tobytes()) for temp_bit_chunk in temp_bit_chunks]

def deinterleave_all(chunks: list[bytes], num_ways: int, word_size: int) -> list[bytes]:
    '''Convenience wrapper for deinterleave() to handle multiple input blobs at once'''
    # Note: A list comprehension like the following would be more pythonic here, but...
    # It's actually harder to read AND oddly slightly less performant (typically .1-.2s at the most)
    # It's not clear why it's slower, but not worth using the list comprehension.
    # Even 2 separated list comprehensions isn't faster...
    # return [d for chunk in chunks for d in deinterleave(chunk, num_ways, word_size)]
    temp_chunks = []
    for chunk in chunks:
        temp_chunks.extend(deinterleave(chunk, num_ways, word_size))
    return temp_chunks

def transform_all(chunks: list[bytes], transform_function: Callable,
    *function_args, **function_kwargs) -> list[bytes]:
    '''Apply an existing transformation function to all objects in chunks'''
    temp_chunks = []
    for chunk in chunks:
        temp_chunks.extend(transform_function(chunk, *function_args, **function_kwargs))
    return temp_chunks

def truncate(contents: bytes, max_length: int) -> bytes:
    '''Remove all content from a blob after max_length bytes'''
    return contents[0:max_length]


def splice_out(contents: bytes, start: int, length: int=None, end: int=None) -> bytearray:
    '''Remove an inner section of a blob'''
    if length is None and end is None:
        raise Exception(
            "Splice out needs a length or end value, but received neither.")
    elif length is not None and end is not None:
        raise Exception(
            "Splice out needs a length or end value, but received both.")
    elif end is None:
        end = start + length

    return bytearray(contents[0:start] + contents[end:len(contents)])


def swap_endian(contents: bytes) -> list[bytearray]:
    '''Swap the a blob from little endian to big or vice versa'''
    new_contents = bytearray(len(contents))
    new_contents[0::2] = contents[1::2]
    new_contents[1::2] = contents[0::2]
    return new_contents


def swap_endian_all(chunks: list[bytes]) -> list[bytearray]:
    '''Convenience wrapper for swap_endian() to handle multiple input blobs at once'''
    temp_chunks = []
    for chunk in chunks:
        temp_chunks.append(swap_endian(chunk))
    return temp_chunks


def bit_shuffle(contents: bytes, word_size_bytes: int, bit_order: list[int]) -> bytearray:
    '''Reshuffle bits in a blob using the specified order'''
    # This is a SLOW operation no matter what.
    # Given an example from SFA1:
    # - Bitwise math: ~17.5s
    # - Native Shuffle, bit map as we go: ~12.1s
    # - Native Shuffle, bit map as we go, but use += instead of extend: ~12.0s
    # - Native Shuffle, outer loop via chunker generator, bit map as we go, but use += instead of extend: ~12.8s
    # - Native Shuffle, bit map append as we go, but use += instead of extend: ~13.8s
    # - Native Shuffle, dump all in bitarray and change in place: ~12.5s
    # - Native Shuffle, dump all in bitarray, use itertools generator, use += instead of extend: ~11.2s
    # - Native Shuffle, dump all in bitarray, use itertools generator, use += and explicit tobytes instead of extend: ~11.2s
    # - Native Shuffle, dump all in bitarray, use itertools generator, don't clean updated_word,use += and explicit tobytes instead of extend: ~10.8s
    # - Full numpy process: ~19.0s
    # - Numpy Shuffle, assuming already numpy array: ~19.0s
    # - Numpy Shuffle, without pack/unpack: ~18.9s
    # - Numpy Shuffle ONLY replacing bitarray: ~37.0s
    # - Numpy Shuffle array: ~30.0s and it still isn't modifying in place
    # Also, things that don't seem to work:
    # - more_itertools.grouper gets int arrays instead of byte arrays
    # Other ideas:
    # - Use a memory view?

    # print(f"before: {hash_helper.get_crc(contents)}")
    # shuffle_start = time.perf_counter()
    def chunker(iterable, size, fillvalue=None):
        args = [iter(iterable)] * size
        return itertools.zip_longest(*args, fillvalue=fillvalue)
    word_size_bits = word_size_bytes*8
    bit_contents = bitarray()
    bit_contents.frombytes(contents)
    new_content = bytearray()
    updated_word = bitarray(word_size_bits*'0')
    for shuffle_word in chunker(bit_contents, word_size_bits):
        # Perform shuffle
        for idx, next_bit in enumerate(bit_order):
            updated_word[idx] = shuffle_word[next_bit]
        # Append to output
        new_content += updated_word.tobytes()
    return new_content


def split_bit_shuffle(contents: bytes, word_size_bytes: int, bit_order: list[int], num_ways: int) -> list[bytearray]:
    '''Reshuffle bits in a blob using the specified order into an array of blobs'''
    new_chunks = []
    for curr_way in range(0, num_ways):
        new_chunks.append(bytearray())

    num_shuffles = len(contents)//word_size_bytes
    shuffle_word = bitarray()
    updated_word = bitarray(word_size_bytes*8*'0')
    for i in range(0, num_shuffles):
        offset = i*word_size_bytes
        shuffle_word.clear()
        shuffle_word.frombytes(contents[offset:offset+word_size_bytes])

        updated_word.setall(0)
        j = 0
        for next_bit in bit_order:
            updated_word[j] = shuffle_word[next_bit]
            j = j + 1

        offset = 0
        split_shuffle_length = word_size_bytes*8//num_ways
        for curr_way in range(0, num_ways):
            end = offset+split_shuffle_length
            new_chunks[curr_way] += updated_word[offset:end]
            offset = end

    return new_chunks


def byte_shuffle(contents: bytes, word_size_bytes: int, byte_order: list[int]) -> bytearray:
    '''Reshuffle bytes in a blob using the specified order'''
    new_content = bytearray()
    num_shuffles = len(contents)//word_size_bytes
    updated_word = bytearray(word_size_bytes)
    for i in range(0, num_shuffles):
        offset = i*word_size_bytes
        shuffle_word = bytearray(contents[offset:offset+word_size_bytes])

        updated_word.clear()
        for idx, next_byte in enumerate(byte_order):
            updated_word[idx] = shuffle_word[next_byte]

        new_content += updated_word
    return new_content



'''Implementation of cfc: Capcom Fighting Collection'''
import logging
import os

from gex.lib.archive import arc
from gex.lib.utils.vendor import capcom
from gex.lib.tasks.basetask import BaseTask
from gex.lib.tasks import helpers

logger = logging.getLogger('gextoolbox')

class CFCTask(BaseTask):
    '''Implements cfc: Capcom Fighting Collection'''
    _task_name = "cfc"
    _title = "Capcom Fighting Collection"
    _details_markdown = '''
This is reverse-engineered based on the CBEYB work from https://web.archive.org/web/20220213232104/http://blog.livedoor.jp/scrap_a/archives/23114141.html.
This was a Japanese set of shell scripts and odd generic operation executables. There is some weird encoding here too.

This script will extract and prep the ROMs. Some per-rom errata are in the notes below. All CRCs are currently mismatched.
'''
    _default_input_folder = helpers.gen_steam_app_default_folder("CAPCOM FIGHTING COLLECTION")
    _input_folder_desc = "CFC Steam folder"

    def get_out_file_info(self):
        '''Return a list of output files'''
        return {
            "files": self._metadata['out']['files'],
            "notes": self._metadata['out']['notes']
        }

    def execute(self, in_dir, out_dir):
        # for each output file entry
        for out_file_entry in self._metadata['out']['files']:
            pkg_name = out_file_entry['in_file']
            # Check the status of it
            if out_file_entry['status'] == 'no-rom':
                logger.info(f"Skipping {pkg_name} - cannot extract...")
            else:
                logger.info(f"Extracting {pkg_name}...")

                # read the matching input file
                in_file_entry = self._metadata['in']['files'][pkg_name]
                loaded_file = self.read_datafile(in_dir, in_file_entry)

                # extract the input file
                arc_contents = arc.extract(loaded_file['contents'])

                # Get the bin entry
                merged_rom_contents = None
                for _, arc_content in arc_contents.items():
                    if arc_content['path'].startswith('bin'):
                        merged_rom_contents = arc_content['contents']

                # run the handler
                handler_func = self.find_handler_func(pkg_name)
                if merged_rom_contents is not None and handler_func is not None:
                    output_contents = handler_func(merged_rom_contents)

                    _ = self.verify_out_file(out_file_entry['filename'], output_contents)

                    with open(os.path.join(out_dir, out_file_entry['filename']), "wb") as out_file:
                        out_file.write(output_contents)
                elif merged_rom_contents is None:
                    logger.warning("Could not find merged rom data in arc.")
                elif handler_func is None:
                    logger.warning("Could not find matching handler function.")
        logger.info("Processing complete.")


    ################################################################################
    # Darkstalkers/Vampire: The Night Warriors                                     #
    ################################################################################

    _vam_gfx_filenames = [
        'vam.13m',
        'vam.14m',
        'vam.15m',
        'vam.16m',
        'vam.17m',
        'vam.18m',
        'vam.19m',
        'vam.20m'
    ]

    _vam_audiocpu_filenames = [
        'vam.01',
        'vam.02'
    ]

    _vam_qsound_filenames = [
        'vam.11m',
        'vam.12m'
    ]

    def _handle_vampj(self, merged_contents):
        func_map = {}

        maincpu_filenames = [
            "vamj.03a",
            "vamj.04b",
            "vamj.05a",
            "vamj.06a",
            "vamj.07a",
            "vamj.08a",
            "vamj.09a",
            "vamj.10a"
        ]
        func_map['maincpu'] = capcom.maincpu_cps2(
            0x40, 0x400000, 8, maincpu_filenames)
        func_map['gfx'] = capcom.gfx_cps2(self._vam_gfx_filenames, helpers.slice_helper(
            0x0800040, length=0x1400000), split=[0x400000, 0x100000])
        func_map['audiocpu'] = capcom.audiocpu_cps2(
            0x1C00040, self._vam_audiocpu_filenames)
        func_map['qsound'] = capcom.qsound_cps2(
            0x1C50040, 0x400000, self._vam_qsound_filenames)

        return helpers.build_rom(merged_contents, func_map)

    def _handle_dstlku(self, merged_contents):
        func_map = {}

        maincpu_filenames = [
            "vamu.03b",
            "vamu.04b",
            "vamu.05b",
            "vamu.06b",
            "vamu.07b",
            "vamu.08b",
            "vamu.09b",
            "vamu.10b"
        ]
        func_map['maincpu'] = capcom.maincpu_cps2(
            0x40, 0x400000, 8, maincpu_filenames)
        func_map['gfx'] = capcom.gfx_cps2(self._vam_gfx_filenames, helpers.slice_helper(
            0x0800040, length=0x1400000), split=[0x400000, 0x100000])
        func_map['audiocpu'] = capcom.audiocpu_cps2(
            0x1C00040, self._vam_audiocpu_filenames)
        func_map['qsound'] = capcom.qsound_cps2(
            0x1C50040, 0x400000, self._vam_qsound_filenames)

        return helpers.build_rom(merged_contents, func_map)

    ################################################################################
    # Vampire Hunter/Night Warriors: Darkstalkers' Revenge                         #
    ################################################################################

    _vph_gfx_filenames = [
        'vph.13m',
        'vph.14m',
        'vph.15m',
        'vph.16m',
        'vph.17m',
        'vph.18m',
        'vph.19m',
        'vph.20m'
    ]

    _vph_audiocpu_filenames = [
        'vph.01',
        'vph.02'
    ]

    _vph_qsound_filenames = [
        'vph.11m',
        'vph.12m'
    ]

    def _handle_vhuntjr2(self, merged_contents):
        func_map = {}

        maincpu_filenames = [
            "vphj.03b",
            "vphj.04a",
            "vphj.05a",
            "vphj.06a",
            "vphj.07a",
            "vphj.08a",
            "vphj.09a",
            "vphj.10a"
        ]
        func_map['maincpu'] = capcom.maincpu_cps2(
            0x40, 0x400000, 8, maincpu_filenames)
        func_map['gfx'] = capcom.gfx_cps2(self._vph_gfx_filenames, helpers.slice_helper(
            0x0800040, length=0x2000000), split=[0x400000, 0x400000])
        func_map['audiocpu'] = capcom.audiocpu_cps2(
            0x2800040, self._vph_audiocpu_filenames)
        func_map['qsound'] = capcom.qsound_cps2(
            0x2850040, 0x400000, self._vph_qsound_filenames)

        return helpers.build_rom(merged_contents, func_map)

    def _handle_nwarru(self, merged_contents):

        maincpu_filenames = [
            "vphu.03f",
            "vphu.04c",
            "vphu.05e",
            "vphu.06c",
            "vphu.07b",
            "vphu.08b",
            "vphu.09b",
            "vphu.10b"
        ]

        func_map = {}
        func_map['maincpu'] = capcom.maincpu_cps2(
            0x40, 0x400000, 8, maincpu_filenames)
        func_map['gfx'] = capcom.gfx_cps2(self._vph_gfx_filenames, helpers.slice_helper(
            0x0800040, length=0x2000000), split=[0x400000, 0x400000])
        func_map['audiocpu'] = capcom.audiocpu_cps2(
            0x2800040, self._vph_audiocpu_filenames)
        func_map['qsound'] = capcom.qsound_cps2(
            0x2850040, 0x400000, self._vph_qsound_filenames)

        return helpers.build_rom(merged_contents, func_map)

    ################################################################################
    # Vampire Savior: The Lord of Vampire                                          #
    ################################################################################

    _vm3_gfx_filenames = [
        'vm3.13m',
        'vm3.14m',
        'vm3.15m',
        'vm3.16m',
        'vm3.17m',
        'vm3.18m',
        'vm3.19m',
        'vm3.20m'
    ]

    _vm3_audiocpu_filenames = [
        'vm3.01',
        'vm3.02'
    ]

    _vm3_qsound_filenames = [
        'vm3.11m',
        'vm3.12m'
    ]

    def _handle_vsavj(self, merged_contents):
        func_map = {}

        maincpu_filenames = [
            "vm3j.03d",
            "vm3j.04d",
            "vm3j.05a",
            "vm3j.06b",
            "vm3j.07b",
            "vm3j.08a",
            "vm3j.09b",
            "vm3j.10b"
        ]
        func_map['maincpu'] = capcom.maincpu_cps2(
            0x40, 0x400000, 8, maincpu_filenames)
        func_map['gfx'] = capcom.gfx_cps2(self._vm3_gfx_filenames, helpers.slice_helper(
            0x0800040, length=0x2000000), split=[0x400000, 0x400000])
        func_map['audiocpu'] = capcom.audiocpu_cps2(
            0x2800040, self._vm3_audiocpu_filenames)
        func_map['qsound'] = capcom.qsound_cps2(
            0x2850040, 0x800000, self._vm3_qsound_filenames)

        return helpers.build_rom(merged_contents, func_map)

    def _handle_vsavu(self, merged_contents):
        func_map = {}

        maincpu_filenames = [
            "vm3u.03d",
            "vm3u.04d",
            "vm3.05a",
            "vm3.06a",
            "vm3.07b",
            "vm3.08a",
            "vm3.09b",
            "vm3.10b"
        ]
        func_map['maincpu'] = capcom.maincpu_cps2(
            0x40, 0x400000, 8, maincpu_filenames)
        func_map['gfx'] = capcom.gfx_cps2(self._vm3_gfx_filenames, helpers.slice_helper(
            0x0800040, length=0x2000000), split=[0x400000, 0x400000])
        func_map['audiocpu'] = capcom.audiocpu_cps2(
            0x2800040, self._vm3_audiocpu_filenames)
        func_map['qsound'] = capcom.qsound_cps2(
            0x2850040, 0x800000, self._vm3_qsound_filenames)

        return helpers.build_rom(merged_contents, func_map)

    ################################################################################
    # Vampire Hunter 2: Darkstalkers Revenge                                       #
    ################################################################################

    _vh2_gfx_filenames = [
        'vh2.13m',
        'vh2.14m',
        'vh2.15m',
        'vh2.16m',
        'vh2.17m',
        'vh2.18m',
        'vh2.19m',
        'vh2.20m'
    ]

    _vh2_audiocpu_filenames = [
        'vh2.01',
        'vh2.02'
    ]

    _vh2_qsound_filenames = [
        'vh2.11m',
        'vh2.12m'
    ]

    def _handle_vhunt2(self, merged_contents):
        func_map = {}

        maincpu_filenames = [
            "vh2j.03a",
            "vh2j.04a",
            "vh2j.05",
            "vh2j.06",
            "vh2j.07",
            "vh2j.08",
            "vh2j.09",
            "vh2j.10"
        ]
        func_map['maincpu'] = capcom.maincpu_cps2(
            0x40, 0x400000, 8, maincpu_filenames)
        func_map['gfx'] = capcom.gfx_cps2(self._vh2_gfx_filenames, helpers.slice_helper(
            0x0800040, length=0x2000000), split=[0x400000, 0x400000])
        func_map['audiocpu'] = capcom.audiocpu_cps2(
            0x2800040, self._vh2_audiocpu_filenames)
        func_map['qsound'] = capcom.qsound_cps2(
            0x2850040, 0x800000, self._vh2_qsound_filenames)

        return helpers.build_rom(merged_contents, func_map)

    ################################################################################
    # Vampire Savior 2: The Lord of Vampire                                        #
    ################################################################################

    _vs2_gfx_filenames = [
        'vs2.13m',
        'vs2.14m',
        'vs2.15m',
        'vs2.16m',
        'vs2.17m',
        'vs2.18m',
        'vs2.19m',
        'vs2.20m'
    ]

    _vs2_audiocpu_filenames = [
        'vs2.01',
        'vs2.02'
    ]

    _vs2_qsound_filenames = [
        'vs2.11m',
        'vs2.12m'
    ]

    def _handle_vsav2(self, merged_contents):
        func_map = {}

        maincpu_filenames = [
            "vs2j.03",
            "vs2j.04",
            "vs2j.05",
            "vs2j.06",
            "vs2j.07",
            "vs2j.08",
            "vs2j.09",
            "vs2j.10"
        ]
        func_map['maincpu'] = capcom.maincpu_cps2(
            0x40, 0x400000, 8, maincpu_filenames)
        func_map['gfx'] = capcom.gfx_cps2(self._vs2_gfx_filenames, helpers.slice_helper(
            0x0800040, length=0x2000000), split=[0x400000, 0x400000])
        func_map['audiocpu'] = capcom.audiocpu_cps2(
            0x2800040, self._vs2_audiocpu_filenames)
        func_map['qsound'] = capcom.qsound_cps2(
            0x2850040, 0x800000, self._vs2_qsound_filenames)

        return helpers.build_rom(merged_contents, func_map)

    ################################################################################
    # Cyberbots: Fullmetal Madness                                                 #
    ################################################################################

    _cybots_gfx_filenames = [
        'cyb.13m',
        'cyb.14m',
        'cyb.15m',
        'cyb.16m',
        'cyb.17m',
        'cyb.18m',
        'cyb.19m',
        'cyb.20m'
    ]

    _cybots_audiocpu_filenames = [
        'cyb.01',
        'cyb.02'
    ]

    _cybots_qsound_filenames = [
        'cyb.11m',
        'cyb.12m'
    ]

    def _handle_cybotsj(self, merged_contents):
        func_map = {}

        maincpu_filenames = [
            "cybj.03",
            "cybj.04",
            "cyb.05",
            "cyb.06",
            "cyb.07",
            "cyb.08",
            "cyb.09",
            "cyb.10"
        ]
        func_map['maincpu'] = capcom.maincpu_cps2(
            0x40, 0x400000, 8, maincpu_filenames)
        func_map['gfx'] = capcom.gfx_cps2(self._cybots_gfx_filenames, helpers.slice_helper(
            0x0800040, length=0x2000000), split=[0x400000, 0x400000])
        func_map['audiocpu'] = capcom.audiocpu_cps2(
            0x2800040, self._cybots_audiocpu_filenames)
        func_map['qsound'] = capcom.qsound_cps2(
            0x2850040, 0x400000, self._cybots_qsound_filenames)

        return helpers.build_rom(merged_contents, func_map)

    def _handle_cybotsu(self, merged_contents):
        func_map = {}

        maincpu_filenames = [
            "cybu.03",
            "cybu.04",
            "cyb.05",
            "cyb.06",
            "cyb.07",
            "cyb.08",
            "cyb.09",
            "cyb.10"
        ]
        func_map['maincpu'] = capcom.maincpu_cps2(
            0x40, 0x400000, 8, maincpu_filenames)
        func_map['gfx'] = capcom.gfx_cps2(self._cybots_gfx_filenames, helpers.slice_helper(
            0x0800040, length=0x2000000), split=[0x400000, 0x400000])
        func_map['audiocpu'] = capcom.audiocpu_cps2(
            0x2800040, self._cybots_audiocpu_filenames)
        func_map['qsound'] = capcom.qsound_cps2(
            0x2850040, 0x400000, self._cybots_qsound_filenames)

        return helpers.build_rom(merged_contents, func_map)

    ################################################################################
    # Super Puzzle Fighter II Turbo                                                #
    ################################################################################

    _pzf_gfx_filenames = [
        'pzf.14m',
        'pzf.16m',
        'pzf.18m',
        'pzf.20m'
    ]

    _pzf_audiocpu_filenames = [
        'pzf.01',
        'pzf.02'
    ]

    _pzf_qsound_filenames = [
        'pzf.11m',
        'pzf.12m'
    ]

    def _handle_spf2xj(self, merged_contents):
        func_map = {}

        maincpu_filenames = [
            "pzfj.03a",
            "pzf.04"
        ]
        func_map['maincpu'] = capcom.maincpu_cps2(
            0x40, 0x100000, 2, maincpu_filenames)
        func_map['gfx'] = capcom.gfx_cps2(self._pzf_gfx_filenames, helpers.slice_helper(
            0x1000040, length=0x400000), split=[0x100000])
        func_map['audiocpu'] = capcom.audiocpu_cps2(
            0x1400040, self._pzf_audiocpu_filenames)
        func_map['qsound'] = capcom.qsound_cps2(
            0x1450040, 0x400000, self._pzf_qsound_filenames)

        return helpers.build_rom(merged_contents, func_map)

    def _handle_spf2tu(self, merged_contents):
        func_map = {}

        maincpu_filenames = [
            "pzfu.03a",
            "pzf.04"
        ]
        func_map['maincpu'] = capcom.maincpu_cps2(
            0x40, 0x100000, 2, maincpu_filenames)
        func_map['gfx'] = capcom.gfx_cps2(self._pzf_gfx_filenames, helpers.slice_helper(
            0x1000040, length=0x400000), split=[0x100000])
        func_map['audiocpu'] = capcom.audiocpu_cps2(
            0x1400040, self._pzf_audiocpu_filenames)
        func_map['qsound'] = capcom.qsound_cps2(
            0x1450040, 0x400000, self._pzf_qsound_filenames)

        return helpers.build_rom(merged_contents, func_map)

    ################################################################################
    # Super Gem Fighter Mini Mix                                                   #
    ################################################################################

    _pcf_gfx_filenames = [
        'pcf.13m',
        'pcf.14m',
        'pcf.15m',
        'pcf.16m',
        'pcf.17m',
        'pcf.18m',
        'pcf.19m',
        'pcf.20m'
    ]

    _pcf_audiocpu_filenames = [
        'pcf.01',
        'pcf.02'
    ]

    _pcf_qsound_filenames = [
        'pcf.11m',
        'pcf.12m'
    ]

    def _handle_pfghtj(self, merged_contents):
        func_map = {}

        maincpu_filenames = [
            "pcfj.03",
            "pcf.04",
            "pcf.05",
            "pcf.06",
            "pcf.07"
        ]
        func_map['maincpu'] = capcom.maincpu_cps2(
            0x40, 0x280000, 5, maincpu_filenames)
        func_map['gfx'] = capcom.gfx_cps2(self._pcf_gfx_filenames, helpers.slice_helper(
            0x0800040, length=0x1400000), split=[0x400000, 0x100000])
        func_map['audiocpu'] = capcom.audiocpu_cps2(
            0x1C00040, self._pcf_audiocpu_filenames)
        func_map['qsound'] = capcom.qsound_cps2(
            0x1C50040, 0x800000, self._pcf_qsound_filenames)

        return helpers.build_rom(merged_contents, func_map)

    def _handle_sgemf(self, merged_contents):
        func_map = {}

        maincpu_filenames = [
            "pcfu.03a",
            "pcf.04"
        ]
        func_map['maincpu'] = capcom.maincpu_cps2(
            0x40, 0x280000, 5, maincpu_filenames)
        func_map['gfx'] = capcom.gfx_cps2(self._pcf_gfx_filenames, helpers.slice_helper(
            0x0800040, length=0x1400000), split=[0x400000, 0x100000])
        func_map['audiocpu'] = capcom.audiocpu_cps2(
            0x1C00040, self._pcf_audiocpu_filenames)
        func_map['qsound'] = capcom.qsound_cps2(
            0x1C50040, 0x800000, self._pcf_qsound_filenames)

        return helpers.build_rom(merged_contents, func_map)

    ################################################################################
    # Hyper Street Fighter II: Anniversary Edition                                 #
    ################################################################################

    _hs2_gfx_filenames = [
        'hs2.13m',
        'hs2.15m',
        'hs2.17m',
        'hs2.19m'
    ]

    _hs2_audiocpu_filenames = [
        'hs2.01',
        'hs2.02'
    ]

    _hs2_qsound_filenames = [
        'hs2.11m'
    ]

    def _handle_hsf2j(self, merged_contents):
        func_map = {}

        maincpu_filenames = [
            "hs2j.03",
            "hs2j.04",
            "hs2j.05",
            "hs2j.06",
            "hs2j.07",
            "hs2j.08",
            "hs2j.09",
            "hs2j.10"
        ]
        func_map['maincpu'] = capcom.maincpu_cps2(
            0x40, 0x400000, 8, maincpu_filenames)
        func_map['gfx'] = capcom.gfx_cps2(self._hs2_gfx_filenames, helpers.slice_helper(
            0x0800040, length=0x2000000), split=[0x800000])
        func_map['audiocpu'] = capcom.audiocpu_cps2(
            0x2800040, self._hs2_audiocpu_filenames)
        func_map['qsound'] = capcom.qsound_cps2(
            0x2850040, 0x800000, self._hs2_qsound_filenames, num_chunks=1)

        return helpers.build_rom(merged_contents, func_map)

    def _handle_hsf2(self, merged_contents):
        func_map = {}

        maincpu_filenames = [
            "hs2u.03",
            "hs2u.04",
            "hs2.05",
            "hs2.06",
            "hs2.07",
            "hs2.08",
            "hs2.09",
            "hs2.10"
        ]
        func_map['maincpu'] = capcom.maincpu_cps2(
            0x40, 0x400000, 8, maincpu_filenames)
        func_map['gfx'] = capcom.gfx_cps2(self._hs2_gfx_filenames, helpers.slice_helper(
            0x0800040, length=0x2000000), split=[0x800000])
        func_map['audiocpu'] = capcom.audiocpu_cps2(
            0x2800040, self._hs2_audiocpu_filenames)
        func_map['qsound'] = capcom.qsound_cps2(
            0x2850040, 0x800000, self._hs2_qsound_filenames, num_chunks=1)

        return helpers.build_rom(merged_contents, func_map)

    ################################################################################
    # Red Earth                                                                    #
    ################################################################################
    # game_90.arc: Warzard (JP)
    # game_91.arc: Red Earth

    # This does not appear to be the old NO-CD release (0.139 would target it!)

    # This is CPS3.
    # The actual ROM is small - only 512KB - but there doesn't appear to be an exact match offhand.
    # The majority of the data is in a CHD - but there doesn't appear to be an exact match offhand.

    # Warzard has a 'JACK' header, but Red Earth has an 'IBIS' header. Likely same format.

    # There are very 'word like' sections that cryptanalysis can be done with.
    # EX: 0x1A490
    #    orig: M:noadnouT:yuT:e
    #  endian: :MondaonTuy:Tue:
    # This section is pretty clearly days of week, date format strings and months of year.
    # It LOOKS like every 4 charcters are backwards
    #     sep: M:no | adno | uT:y | uT:e
    #  4 char: on:M | onda | y:Tu | e:Tue
    #  merged: on:Monday:Tue:Tue
    # Doing this to the whole file gets us a lot of readable text (Days/Months/Error Strings)...
    # But that readable text isn't in the romfile or compressed CHD.
    # Did they pre-decompress the CHD for performance?

    # There are CD001 headers - is this an ISO IMAGE?!
    # https://en.wikipedia.org/wiki/ISO_9660#Specifications
    # 2 Tracks:
    #  0x1A1CF: hCD001� - type 68, ver 0 (invalid)
    #  0x1D80C: hCD001� - type 68, ver 0 (invalid)
    # There would be 32,768 bytes (0x8000) in a 'system area' before the first CD001 header
    # This would mean the image starts at 0x1A1CF - 0x8000 = 0x121CF
    # This is 0x0 - 0x121CF is NOT big enough to fit the 0x80000 ROM. Nor is 0x0 - 0x1A1CF.
    # More importantly - this isn't a valid ISO 9660
    # The volume versions are invalid, there's not a primary or terminator...

    # 0x0 - 0x40 is the IBIS/JACK header.
    # 0x40 - 0x1FF40 matches 0x0 - 0x1FF00 in the dump from MAME.
    # After that, however, the files differ substantially - the vast majority of the file from CFC is blank.

    # After 0x80040 there needs to be the equivalent of the .CHD file for the game. (TLDR, CPS3 games had a ROM and a CD-ROM, and MAME uses the .CHD to represent the CDROM.)

    # However, even a larger dump saveo warzard-full.bin,0,2080000 looks like it has the CHD header, etc.

    # def _handle_warzard(self, merged_contents):
    #     out_files = []
    #     func_map = {}

    #     def endian(contents):
    #         contents = transforms.byte_shuffle(contents, 4, [3, 2, 1, 0])
    #         return { 'endian.bin': contents }

    #     func_map['endian'] = endian
    #     return 
    #         {'filename': 'warzard.zip', 'contents': helpers.build_rom(merged_contents, func_map)}
    #     )
    #     return out_files

    # def _handle_redearth(self, merged_contents):
    #     out_files = []
    #     func_map = {}

    #     def endian(contents):
    #         contents = transforms.byte_shuffle(contents, 4, [3, 2, 1, 0])
    #         return { 'endian.bin': contents }

    #     func_map['endian'] = endian
    #     return 
    #         {'filename': 'redearth.zip', 'contents': helpers.build_rom(merged_contents, func_map)}
    #     )
    #     return out_files



'''Convenience functions for comparing two files/archives'''
from enum import Enum
import logging
import os
from gex.lib.utils.blob import hash as hash_helper
from gex.lib.archive import zip as zip_lib

logger = logging.getLogger('gextoolbox')

class VerifyErrors(Enum):
    MISSING_FILE = 1
    EXTRA_FILE = 2
    SIZE_MISMATCH = 3
    CHECKSUM_MISMATCH = 4
    VERSION_MISMATCH = 5

def verify_via_metadata(contents, verify_metadata):
    '''Wrapper to apply any verify test; returns errors, than any verification notes (such as matched version)'''
    if verify_metadata['type'] == "zip":
        return verify_zip(contents, verify_metadata['entries']), None
    elif verify_metadata['type'] == "sha":
        return verify_sha(contents, verify_metadata['sha'], verify_metadata['size']), None
    elif verify_metadata['type'] == "crc":
        return verify_crc(contents, verify_metadata['crc'], verify_metadata['size']), None
    elif verify_metadata['type'] == "versions":
        return verify_versions(contents, verify_metadata['versions'])


def verify_versions(contents, versions):
    all_version_errors = {}
    for version_name, version_verify_metadata in versions.items():
        all_version_errors[version_name] = verify_via_metadata(contents, version_verify_metadata)

    closest_version = min(all_version_errors, key=all_version_errors.get)
    version_errors = all_version_errors[closest_version]
    return version_errors, closest_version 


def verify_zip(zip_data, known_entries):
    errors = []

    zip_metas = zip_lib.get_metadata(zip_data)

    # Ensure the file name lists are the same
    real_filenames = set(zip_metas.keys())
    expected_filenames = set(known_entries.keys())
    if real_filenames != expected_filenames:
        missing_files = expected_filenames.difference(real_filenames)
        for missing_file in missing_files:
            errors.append({VerifyErrors.MISSING_FILE, missing_file})
        extra_files = real_filenames.difference(expected_filenames)
        for extra_file in extra_files:
            errors.append({VerifyErrors.EXTRA_FILE, extra_file})

    # Compare each file size/CRC
    check_files = real_filenames.intersection(expected_filenames)
    for inner_filename in check_files:
        zip_meta = zip_metas.get(inner_filename)
        verify_entry = known_entries[inner_filename]
        if zip_meta['crc'] != verify_entry['crc']:
            errors.append((VerifyErrors.CHECKSUM_MISMATCH, {zip_meta['crc'], verify_entry['crc'], "CRC", inner_filename}))
        if zip_meta['size'] != verify_entry['size']:
            errors.append((VerifyErrors.SIZE_MISMATCH, {zip_meta['size'], verify_entry['size'], inner_filename}))
            
    return errors

def verify_crc(contents, known_crc, known_size):
    size = len(contents)
    crc = hash_helper.get_crc(contents)
    known_crc = hex(int(known_crc, base=16))
    errors = []
    if known_size != size:
        errors.append((VerifyErrors.SIZE_MISMATCH, {size, known_size}))

    if known_crc != crc:
        errors.append((VerifyErrors.CHECKSUM_MISMATCH, {crc, known_crc, "CRC"}))

    return errors

def verify_sha(contents, known_sha, known_size):
    size = len(contents)
    sha = hash_helper.get_sha1(contents)
    errors = []
    if known_size != size:
        errors.append((VerifyErrors.SIZE_MISMATCH, {size, known_size}))

    if known_sha != sha:
        errors.append((VerifyErrors.CHECKSUM_MISMATCH, {sha, known_sha, "SHA-1"}))

    return errors

def print_errors(errors, ref):
    logger.info(f"Verification of {ref} failed!")
    for error in errors:
        message = ""
        type = error[0]
        data = error[1]
        if type is VerifyErrors.MISSING_FILE:
            message = f"File {data[0]} is missing!"
        elif type is VerifyErrors.EXTRA_FILE:
            message = f"File {data[0]} is extraneous!"
        elif type is VerifyErrors.SIZE_MISMATCH:
            if len(data) == 2:
                message = f"Expected size {data[0]}, found {data[1]}!"
            elif len(data) == 3:
                message = f"Expected size {data[0]} for {data[2]}, found {data[1]}!"
        elif type is VerifyErrors.CHECKSUM_MISMATCH:
            if len(data) == 3:
                message = f"Expected {data[2]} {data[0]}, found {data[1]}!"
            elif len(data) == 4:
                message = f"Expected {data[2]} {data[0]} for {data[3]}, found {data[1]}!"
        logger.info(f"- {message}")


    def verify_out_file(self, file_name, contents):
        '''Verify an output file using the method specified in the metadata'''
        # Find out_file entry
        out_file = next((x for x in self._metadata['out']['files'] if x['filename'] == file_name), None)
        if out_file is None:
            logger.info(f"Could not find entry to verify {file_name}")
            return None

        if out_file['status'] == 'partial' or out_file['status'] == 'no-rom':
            logger.info(f"Verification not available for partial ROM {file_name}")
            return None

        # Get verify object
        verify_obj = out_file.get('verify')
        if verify_obj is None:
            logger.info(f"Verification data not available for ROM {file_name}")
            return None

        # Check verify type
        if verify_obj['type'] == 'crc':
            crc = hash_helper.get_crc(contents)[2:].upper().rjust(8, "0")
            if len(contents) != verify_obj['size']:
                logger.info(f"Could NOT verify {file_name}: File size doesn't match!")
                return False
            elif crc != verify_obj['crc']:
                logger.info(f"Could NOT verify {file_name}: CRC doesn't match!")
                return False
            else:
                logger.info(f"Verified {file_name}.")
                return True
        elif verify_obj['type'] == 'zip':
            zip_metas = zip_lib.get_metadata(contents)
            # Ensure the file name lists are the same
            real_filenames = set(zip_metas.keys())
            expected_filenames = set(verify_obj['entries'].keys())
            if real_filenames != expected_filenames:
                logger.info(f"Could NOT verify {file_name}: File lists don't match!")
                return False
            # Compare file size/CRC
            for inner_filename, zip_meta in zip_metas.items():
                verify_entry = verify_obj['entries'][inner_filename]
                if zip_meta['crc'] != verify_entry['crc'] or zip_meta['size'] != verify_entry['size'] :
                    logger.info(f"Could NOT verify {file_name}: {inner_filename} should be {verify_entry['crc']} at {verify_entry['size']} bytes, found {zip_meta['crc']} at {zip_meta['size']}")
                    return False
            logger.info(f"Verified {file_name}.")
            return True
        else:
            logger.info(f"Unknown verify type: {verify_obj['type']}")
            return None



'''Implementation of basic splice task'''
import logging
import os
from gex.lib.tasks.basetask import BaseTask

logger = logging.getLogger('gextoolbox')

class SpliceTask(BaseTask):
    '''Implements basic splice task'''

    def get_out_file_info(self):
        '''Return a list of output files'''
        return {
            "files": self._metadata['out']['files'],
            "notes": self._metadata['out']['notes']
        }

    def execute(self, in_dir, out_dir):
        '''Splice out the ROM files'''
        resolved_file = self.read_datafile(in_dir, self._metadata['in']['files']['source'])
        source_data = resolved_file['contents']

        extractable_roms = [x for x in self._metadata['out']['files'] if x['status'] != 'no-rom']
        for file_meta in extractable_roms:
            logger.info(f"Extracting {file_meta['game']}...")
            game_data = bytearray()
            if 'header' in file_meta:
                game_data.extend(bytearray.fromhex(file_meta['header']))

            for section in file_meta['sections'].values():
                start = int(section['start'], 16)
                length = int(section['length'], 16)
                game_data.extend(source_data[start:start+length])

            filename = file_meta['filename']
            _ = self.verify_out_file(filename, game_data)
            out_path = os.path.join(out_dir, filename)
            with open(out_path, "wb") as out_file:
                out_file.write(game_data)

        logger.info("Processing complete.")



'''Containes the BaseTask class, the base class for tasks'''

import copy
import json
import logging
import os
from gex.lib.utils.blob import hash as hash_helper
from gex.lib.archive import zip as zip_lib
from gex.lib.utils.verify import verify_via_metadata

logger = logging.getLogger('gextoolbox')

class BaseTask:
    '''Base class for rom extraction tasks'''
    _task_name = None
    _title = None
    _details_markdown = None
    _default_input_folder = None
    _input_folder_desc = None
    _out_file_list = []
    _out_file_notes = {}
    _prop_info = {}
    _props = {}
    _metadata = None

    def __init__(self):
        for prop_key, prop_value in self._prop_info.items():
            self._props[prop_key] = prop_value['default']

    def read_task_metadata(self):
        '''Read the JSON-based metadata for a task'''
        basetask_path = os.path.dirname(__file__)
        metadata_path = os.path.join(basetask_path, 'impl', self._task_name, 'metadata.json')
        if os.path.exists(metadata_path):
            with open(metadata_path, 'r', encoding="UTF-8") as metadata_file:
                self._metadata = json.load(metadata_file)
        if self._metadata is None:
            logger.warning(f"Could not find metadata.json for {self._task_name}")

    # Reading data files
    # - Multiple mapping types
    #   - One-to-one: Sometimes, one file is one game
    #   - One-to-many: Sometimes, one input file is multiple games
    #   - Many-to-one: Sometimes, we need multiple input files for one game
    #   - Many-to-many: Sometimes, both are true: for examplen read these 5 ROM types, get multiple variants of the same game out of these files
    # - Files can be very large - some games, for example, have single files above 8 GB!
    # - Some are straight copies, some are unpacking complex formats
    # - Disk IO is expensive, but... it could be worse
    # There's no one size fits all.

    def read_all_datafiles(self, in_dir):
        '''Read all data files as defined in the metadata - mostly useful for smaller collections with many-to-many relationships'''
        # We read all files at once generally because there are a few cases where we cross-read them to fill out missing files
        in_files = {}
        for file_ref, file_metadata in self._metadata['in']['files'].items():
            data_file = self.read_datafile(in_dir, file_metadata)
            in_files[file_ref] = data_file
        return in_files

    def read_datafile(self, in_dir, file_metadata):
        '''Read a specific data file as defined in the metadata, including CRC/Size check/version identification'''
        rel_path = file_metadata.get('rel_path') or []
        data_path = os.path.join(in_dir, *rel_path, file_metadata['filename'])
        if os.path.exists(data_path):
            # consider switching this to mmap bytearray/readablefile memory-mapped file hybrids to improve memory usage
            with open(data_path, 'rb') as data_file:
                contents = data_file.read()

            size = len(contents)
            crc = hash_helper.get_crc(contents)
            sha = hash_helper.get_sha1(contents)
            logging.debug(f"Read {data_path} with size {size}, crc {crc} and sha {sha}")

            versions = file_metadata.get("versions") or []
            file_ver_tag = None
            for version_tag, version_meta in versions.items():
                # Check size first if we have it
                if 'size' in version_meta:
                    if version_meta['size'] != size:
                        continue

                if 'sha' in version_meta:
                    if hex(int(version_meta['sha'], base=16)) == sha:
                        file_ver_tag = version_tag
                        break
                    else:
                        continue
                
                if 'crc' in version_meta:
                    if hex(int(version_meta['crc'], base=16)) == crc:
                        file_ver_tag = version_tag
                        break
                    else:
                        continue


            if file_ver_tag:
                logging.debug(f"Identified {data_path} as version {file_ver_tag}")
            else:
                logging.warning(f"Could not identify version of {data_path}!")

            return {
                "name": os.path.basename(data_path),
                "fullpath": data_path,
                "contents": contents,
                "version": file_ver_tag
            }
        else:
            logging.warning(f"Could not find file at {data_path}!")
            return None

    def verify_out_file(self, file_name, contents):
        '''Verify an output file using the method specified in the metadata'''
        # Find out_file entry
        out_file = next((x for x in self._metadata['out']['files'] if x['filename'] == file_name), None)
        if out_file is None:
            logger.info(f"Could not find entry to verify {file_name}")
            return None

        if out_file['status'] == 'partial' or out_file['status'] == 'no-rom':
            logger.info(f"Verification not available for partial ROM {file_name}")
            return None

        # Get verify object
        verify_obj = out_file.get('verify')
        if verify_obj is None:
            logger.info(f"Verification data not available for ROM {file_name}")
            return None

        errors, version = verify_via_metadata(contents, verify_obj)

        if not errors:
            if version:
                return version
            else:
                return True

        return errors

    def set_props(self, in_props):
        '''Set any additional props for this task'''
        for value in in_props:
            print(value)
            if '=' in value:
                key, value = value.split('=')
            else:
                key = value
            self._props[key] = value

    def get_out_file_info(self):
        '''Return a list of output files'''
        return {
            "files": copy.deepcopy(self._out_file_list),
            "notes": copy.deepcopy(self._out_file_notes)
        }

    def get_prop_info(self):
        '''Return a list of available props'''
        return copy.deepcopy(self._prop_info)

    def get_task_name(self):
        '''Get the task name, a short string for referring to the task'''
        return self._task_name

    def get_details_markdown(self):
        '''Get markdown for the details/help docs for this task'''
        return self._details_markdown

    def get_title(self):
        '''Get the nicely formatted name of this task'''
        return self._title

    def get_default_input_folder(self):
        '''Get the default input folder for this task'''
        return self._default_input_folder

    def get_input_folder_description(self):
        '''Get a short description of what the input folder should refer to'''
        return self._input_folder_desc

    def find_handler_func(self, package_name):
        '''Utility function to look for a sub-package handler in a task'''
        if f'_handle_{package_name}' in dir(self):
            return getattr(self, f'_handle_{package_name}')
        return None

    def execute(self, in_dir, out_dir):
        '''Main implementation call for the extraction task'''



'''Implementation of spp_columnsiii: Sega Puzzle Pack Columns III'''
import logging
import os
from gex.lib.archive.kvq import extract
from gex.lib.tasks.basetask import BaseTask

logger = logging.getLogger('gextoolbox')

ENCODE_STRING_BYTES = bytearray('Encoded for KGen Ultra / Sega Puzzle Pack / Snake KML 1999! ', 'ascii')

class SSPTask(BaseTask):
    '''Implements spp_columnsiii: Sega Puzzle Pack Columns III'''
    _task_name = "spp_columnsiii"
    _title = "Sega Puzzle Pack Columns III"
    _details_markdown = '''
Based on: https://github.com/zZeck/SegaSmashPackPCUtils

These ROMs are pulled out of kvq files. 
    '''
    _default_input_folder = r"C:\Sega\Columns III"
    _input_folder_desc = "Sega Puzzle Pack Columns III folder"
    _out_file_notes = {}

    def get_out_file_info(self):
        '''Return a list of output files'''
        return {
            "files": self._metadata['out']['files'],
            "notes": self._metadata['out']['notes']
        }
    def execute(self, in_dir, out_dir):
        for file_metadata in self._metadata['in']['files'].values():
            resolved_file = self.read_datafile(in_dir, file_metadata)
            out_file_entry = [x for x in self._metadata['out']['files'] if x['game'] == file_metadata['copy_to']][0]
            rom_bytes = extract(resolved_file['contents'], ENCODE_STRING_BYTES)
            filename = out_file_entry['filename']
            _ = self.verify_out_file(filename, rom_bytes)
            out_path = os.path.join(out_dir, filename)
            with open(out_path, "wb") as out_file:
                out_file.write(rom_bytes)
        logger.info("Processing complete.")



'''Implementation of snk40: SNK 40th Anniversary Collection'''
import glob
import logging
import os
from gex.lib.contrib.bputil import BPListReader
from gex.lib.tasks.basetask import BaseTask
from gex.lib.tasks import helpers
from gex.lib.tasks.impl.snk40 import nes, arcadedlc, arcademain, arcadepatch

logger = logging.getLogger('gextoolbox')

class SNK40thAnniversaryCollectionTask(BaseTask):
    '''Implements snk40: SNK 40th Anniversary Collection'''
    _task_name = "snk40"
    _title = "SNK 40th Anniversary Collection"
    _details_markdown = '''
Based on:
- https://gitlab.com/vaiski/romextract/-/blob/master/scripts/STEAM-865940.sh
- https://github.com/lioneltrs/buildROM
    '''
    _default_input_folder = helpers.gen_steam_app_default_folder("SNK 40th Anniversary Collection")
    _input_folder_desc = "SNK 40th install folder"
    _prop_info = {
        "include-arcade": {
            "description": "Include the fully-formed arcade games that are in SNK 40th",
            "default": True,
            "type": "Boolean"
        },
        "include-nes": {
            "description": "Include the NES ports that are included in SNK 40th",
            "default": True,
            "type": "Boolean"
        }
    }

    def get_out_file_info(self):
        '''Return a list of output files'''
        return {
            "files": self._metadata['out']['files'],
            "notes": self._metadata['out']['notes']
        }

    def execute(self, in_dir, out_dir):
        bundle_contents = {}
        for file_ref, file_metadata in self._metadata['in']['files'].items():
            resolved_file = self.read_datafile(in_dir, file_metadata)
            reader = BPListReader(resolved_file['contents'])
            parsed = reader.parse()
            bundle_contents[file_ref] = parsed

        out_files = []

        if self._props.get('include-nes'):
            out_files.extend(nes.extract(bundle_contents))
        if self._props.get('include-arcade'):
            out_files.extend(arcademain.extract(bundle_contents))
            out_files.extend(arcadedlc.extract(bundle_contents))
            out_files.extend(arcadepatch.extract(bundle_contents))

        if out_files:
            for out_file_entry in out_files:
                filename = out_file_entry['filename']
                _ = self.verify_out_file(filename, out_file_entry ['contents'])
                out_path = os.path.join(out_dir, filename)
                with open(out_path, "wb") as out_file:
                    logger.info(f"Writing {filename}...")
                    out_file.write(out_file_entry['contents'])

        logger.info("Processing complete.")



'''Extraction code for Arcade ROMs from SNK40 Main Bundle'''
import logging
from gex.lib.tasks import helpers
from gex.lib.utils.blob import transforms
from gex.lib.tasks.impl.snk40 import utils

logger = logging.getLogger('gextoolbox')

def extract(bundle_contents):
    '''Extract Arcade ROMs from main bundle'''
    out_files = []
    contents = bundle_contents['main']
    out_files.extend(_handle_prehisle(contents))
    out_files.extend(_handle_streetsm(contents))
    out_files.extend(_handle_ikari3(contents))
    out_files.extend(_handle_vanguard(contents))
    out_files.extend(_handle_pow(contents))
    out_files.extend(_handle_psycho(contents))
    out_files.extend(_handle_victoryroad(contents, bundle_contents['patch']))
    out_files.extend(_handle_ikari(contents))
    out_files.extend(_handle_guerilla(contents))
    out_files.extend(_handle_tnk3(contents))
    out_files.extend(_handle_aso(contents))
    out_files.extend(_handle_athena(contents))
    return out_files

def _handle_prehisle(bundle_contents):
    # PREHISLE Common
    func_map = {}
    out_files = []
    func_map['z80'] = helpers.name_file_helper("PrehistoricIsleIn1930.z80", "gt1.1")
    func_map['tx'] = helpers.name_file_helper("PrehistoricIsleIn1930.w.alpha.rom", "gt15.b15")
    func_map['bgrom'] = helpers.name_file_helper("PrehistoricIsleIn1930.bg.rom", "pi8914.b14")
    func_map['fgrom'] = helpers.name_file_helper("PrehistoricIsleIn1930.fg.rom", "pi8916.h16")
    sprite_file_map = {
        "pi8910.k14": 524288,
        "gt5.5": 131072
    }
    func_map['sprites'] = helpers.custom_split_helper(
        "PrehistoricIsleIn1930.sprites.rom", sprite_file_map)
    func_map['bgmap'] = helpers.name_file_helper("PrehistoricIsleIn1930.bg.map", "gt11.11")
    func_map['samples'] = helpers.name_file_helper("PrehistoricIsleIn1930.samples", "gt4.4")
    logger.info("Processing PREHISLE common files...")
    common_file_map = helpers.process_rom_files(bundle_contents, func_map)

    def prehisle_maincpu(in_file_name, filenames):
        def maincpu(in_files):
            contents = in_files[in_file_name]
            chunks = transforms.deinterleave(contents, num_ways=2, word_size=1)
            return dict(zip(filenames, chunks))
        return maincpu

    # GENSITOU / PREHISLEJ
    func_map = {}
    maincpu_filenames = [
        "gt-j2.2h",
        "gt-j3.3h"
    ]
    func_map['maincpu'] = prehisle_maincpu('PrehistoricIsleIn1930.j.68k', maincpu_filenames)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("gensitou.zip", bundle_contents, func_map))

    # PREHISLEU
    func_map = {}
    maincpu_filenames = [
        "gt-u2.2h",
        "gt-u3.3h"
    ]
    func_map['maincpu'] = prehisle_maincpu('PrehistoricIsleIn1930.u.68k', maincpu_filenames)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("prehisleu.zip", bundle_contents, func_map))

    # PREHISLE
    func_map = {}
    maincpu_filenames = [
        "gt-e2.2h",
        "gt-e3.3h"
    ]
    func_map['maincpu'] = prehisle_maincpu('PrehistoricIsleIn1930.w.68k', maincpu_filenames)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("prehisle.zip", bundle_contents, func_map))
    return out_files

def _handle_streetsm(bundle_contents):
    # STREETS Common
    func_map = {}
    out_files = []
    func_map['soundcpu'] = helpers.name_file_helper("streetsm.soundcpu", "s2-5.16c")
    func_map['upd'] = helpers.name_file_helper("streetsm.upd", "s2-6.18d")
    def streets_gfx2(in_file_name, filenames):
        def gfx2(in_files):
            contents = in_files[in_file_name]
            chunks = transforms.equal_split(contents, 7)
            del chunks[3] # Middle element is junk
            return dict(zip(filenames, chunks))
        return gfx2
    gfx2_filenames = [
        "stsmart.900",
        "stsmart.902",
        "stsmart.904",
        "stsmart.901",
        "stsmart.903",
        "stsmart.905",
    ]
    func_map['gfx2'] = streets_gfx2('streetsm.gfx2', gfx2_filenames)

    logger.info("Processing STREETS common files...")
    common_file_map = helpers.process_rom_files(bundle_contents, func_map)

    # STREETS1 GFX Common
    func_map = {}
    gfx1_filenames = [
        "s2-7.15l",
        "s2-8.15m"
    ]
    func_map['gfx1'] = helpers.equal_split_helper('streetsm1.gfx1', gfx1_filenames)
    logger.info("Processing STREETS1 GFX common files...")
    gfx1_file_map = helpers.process_rom_files(bundle_contents, func_map)

    def streets_maincpu(in_file_name, filenames):
        def maincpu(in_files):
            contents = in_files[in_file_name]
            chunks = transforms.deinterleave(contents, num_ways=2, word_size=1)
            return dict(zip(filenames, chunks))
        return maincpu

    # STREETSM
    func_map = {}
    maincpu_filenames = [
        "s2-1ver2.14h",
        "s2-2ver2.14k"
    ]
    func_map['maincpu'] = streets_maincpu('streetsm.maincpu', maincpu_filenames)
    gfx1_filenames = [
        "s2-9.25l",
        "s2-10.25m"
    ]
    func_map['gfx1'] = helpers.equal_split_helper('streetsm.gfx1', gfx1_filenames)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("streetsm.zip", bundle_contents, func_map))

    # STREETSM1
    func_map = {}
    maincpu_filenames = [
        "s2-1ver1.9c",
        "s2-2ver1.10c"
    ]
    func_map['maincpu'] = streets_maincpu('streetsm1.maincpu', maincpu_filenames)
    func_map['gfx1'] = helpers.existing_files_helper(gfx1_file_map)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("streetsm1.zip", bundle_contents, func_map))

    # STREETSMJ
    func_map = {}
    maincpu_filenames = [
        "s2v1j_01.bin",
        "s2v1j_02.bin"
    ]
    func_map['maincpu'] = streets_maincpu('streetsmj.maincpu', maincpu_filenames)
    func_map['gfx1'] = helpers.existing_files_helper(gfx1_file_map)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("streetsmj.zip", bundle_contents, func_map))

    # STREETSMW
    func_map = {}
    maincpu_filenames = [
        "s-smart1.bin",
        "s-smart2.bin"
    ]
    func_map['maincpu'] = streets_maincpu('streetsmw.maincpu', maincpu_filenames)
    func_map['gfx1'] = helpers.existing_files_helper(gfx1_file_map)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("streetsmw.zip", bundle_contents, func_map))
    return out_files

def _handle_ikari3(bundle_contents):
    # IKARI3 Common
    func_map = {}
    out_files = []
    func_map['upd'] = helpers.name_file_helper("ikari3.upd", "ik3-6.18e")
    user_filenames = [
        "ik3-1.c8",
        "ik3-4.c12"
    ]
    func_map['user'] = helpers.deinterleave_helper('ikari3.user1',
        user_filenames, num_ways=2, word_size=1)

    logger.info("Processing IKARI3 common files...")
    common_file_map = helpers.process_rom_files(bundle_contents, func_map)

    # IKARI3 GFX1 Common
    func_map = {}
    gfx1_filenames = [
        "ik3-7.16l",
        "ik3-8.16m"
    ]
    func_map['gfx1'] = helpers.equal_split_helper('ikari3.gfx1', gfx1_filenames)
    logger.info("Processing IKARI3 GFX1 common files...")
    gfx1_file_map = helpers.process_rom_files(bundle_contents, func_map)

    # IKARI3 GFX2 Common
    func_map = {}
    gfx2_filenames = [
        "ik3-23.bin",
        "ik3-13.bin",
        "ik3-22.bin",
        "ik3-12.bin",
        "ik3-21.bin",
        "ik3-11.bin",
        "ik3-20.bin",
        "ik3-10.bin",
        "ik3-19.bin",
        "ik3-9.bin",
        "ik3-14.bin",
        "ik3-24.bin",
        "ik3-15.bin",
        "ik3-25.bin",
        "ik3-16.bin",
        "ik3-26.bin",
        "ik3-17.bin",
        "ik3-27.bin",
        "ik3-18.bin",
        "ik3-28.bin"
    ]
    def ikari3_gfx2_common(in_file_name, filenames):
        def gfx2(in_files):
            contents = in_files[in_file_name]
            contents = transforms.splice_out(contents, start = 1310720, end = 2097152)
            chunks = transforms.equal_split(contents, 10)
            chunks = transforms.deinterleave_all(chunks, num_ways=2, word_size=1)
            return dict(zip(filenames, chunks))
        return gfx2
    func_map['gfx2'] = ikari3_gfx2_common('ikari3.gfx2', gfx2_filenames)
    logger.info("Processing IKARI3 GFX2 common files...")
    gfx2_file_map = helpers.process_rom_files(bundle_contents, func_map)

    # IKARI3
    func_map = {}
    maincpu_filenames = [
        "ik3-2-ver1.c10",
        "ik3-3-ver1.c9"
    ]
    func_map['maincpu'] = helpers.deinterleave_helper('ikari3.maincpu',
        maincpu_filenames, num_ways=2, word_size=1)
    func_map['gfx1'] = helpers.existing_files_helper(gfx1_file_map)
    func_map['gfx2'] = helpers.existing_files_helper(gfx2_file_map)
    func_map['soundcpu'] = helpers.name_file_helper("ikari3.soundcpu", "ik3-5.16d")
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("ikari3.zip", bundle_contents, func_map))

    # IKARI3J
    func_map = {}
    maincpu_filenames = [
        "ik3-2-j.c10",
        "ik3-3-j.c9"
    ]
    func_map['maincpu'] = helpers.deinterleave_helper('ikari3j.maincpu',
        maincpu_filenames, num_ways=2, word_size=1)
    func_map['gfx1'] = helpers.existing_files_helper(gfx1_file_map)
    func_map['gfx2'] = helpers.existing_files_helper(gfx2_file_map)
    func_map['soundcpu'] = helpers.name_file_helper("ikari3.soundcpu", "ik3-5.16d")
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("ikari3j.zip", bundle_contents, func_map))

    # IKARI3K
    func_map = {}
    maincpu_filenames = [
        "ik3-2k.c10",
        "ik3-3k.c9"
    ]
    func_map['maincpu'] = helpers.deinterleave_helper('ikari3k.maincpu',
        maincpu_filenames, num_ways=2, word_size=1)
    gfx1_filenames = [
        "ik3-7k.16l",
        "ik3-8k.16m"
    ]
    func_map['gfx1'] = helpers.equal_split_helper('ikari3k.gfx1', gfx1_filenames)
    gfx2_filenames = [
        "ikari-880d_t53.d2",
        "ikari-880c_t54.c2",
        "ik12.d1",
        "ik11.c1",
        "ikari-880d_t52.b2",
        "ikari-880c_t51.a2",
        "ik10.b1",
        "ik9.a1",
    ]
    def ikari3k_gfx2(in_file_name, filenames):
        def gfx2(in_files):
            contents = in_files[in_file_name]
            contents = transforms.splice_out(contents, start = 1310720, end = 2097152)
            chunks = transforms.custom_split(contents,
                [1024*1024, 1024*256, 1024*1024, 1024*256])
            chunks = transforms.deinterleave_all(chunks, num_ways=2, word_size=1)
            return dict(zip(filenames, chunks))
        return gfx2
    func_map['gfx2'] = ikari3k_gfx2('ikari3.gfx2', gfx2_filenames)
    func_map['soundcpu'] = helpers.name_file_helper("ikari3.soundcpu", "ik3-5.16d")
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("ikari3k.zip", bundle_contents, func_map))

    # IKARI3U
    func_map = {}
    maincpu_filenames = [
        "ik3-2.c10",
        "ik3-3.c9"
    ]
    func_map['maincpu'] = helpers.deinterleave_helper('ikari3u.maincpu',
        maincpu_filenames, num_ways=2, word_size=1)
    func_map['gfx1'] = helpers.existing_files_helper(gfx1_file_map)
    func_map['gfx2'] = helpers.existing_files_helper(gfx2_file_map)
    func_map['soundcpu'] = helpers.name_file_helper("ikari3.soundcpu", "ik3-5.15d")
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("ikari3u.zip", bundle_contents, func_map))

    return out_files

def _handle_vanguard(bundle_contents):
    # VANGUARD Common
    func_map = {}
    out_files = []

    gfx1_filenames = [
        "sk5_ic50.bin",
        "sk5_ic51.bin"
    ]
    func_map['gfx1'] = helpers.equal_split_helper('vanguard.gfx1', gfx1_filenames)

    prom_filenames = [
        "sk5_ic7.bin",
        "sk5_ic6.bin"
    ]
    func_map['prom'] = helpers.equal_split_helper('vanguard.proms', prom_filenames)

    snk6502_filenames = [
        "sk4_ic51.bin",
        "sk4_ic52.bin"
    ]
    func_map['snk6502'] = helpers.equal_split_helper('vanguard.snk6502', snk6502_filenames)

    speech_file_names = [
        "sk6_ic07.bin",
        "sk6_ic08.bin",
        "sk6_ic11.bin"
    ]
    def vanguard_speech(in_file_name, filenames):
        def speech(in_files):
            contents = in_files[in_file_name]
            contents = contents[16384:]
            chunks = transforms.equal_split(contents, len(filenames))
            return dict(zip(filenames, chunks))
        return speech
    func_map['speech'] = vanguard_speech("vanguard.speech", speech_file_names)
    logger.info("Processing VANGUARD common files...")
    common_file_map = helpers.process_rom_files(bundle_contents, func_map)

    # VANGUARD
    func_map = {}
    maincpu_filenames = [
        "sk4_ic07.bin",
        "sk4_ic08.bin",
        "sk4_ic09.bin",
        "sk4_ic10.bin",
        "sk4_ic13.bin",
        "sk4_ic14.bin",
        "sk4_ic15.bin",
        "sk4_ic16.bin"
    ]
    func_map['maincpu'] = helpers.equal_split_helper('vanguard.maincpu', maincpu_filenames)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("vanguard.zip", bundle_contents, func_map))

    # VANGUARDC
    func_map = {}
    maincpu_filenames = [
        "sk4_ic07.bin",
        "sk4_ic08.bin",
        "sk4_ic09.bin",
        "4",
        "5",
        "sk4_ic14.bin",
        "sk4_ic15.bin",
        "8"
    ]
    func_map['maincpu'] = helpers.equal_split_helper('vanguardc.maincpu', maincpu_filenames)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("vanguardc.zip", bundle_contents, func_map))

    # VANGUARDJ
    func_map = {}
    maincpu_filenames = [
        "sk4_ic07.bin",
        "sk4_ic08.bin",
        "sk4_ic09.bin",
        "vgj4ic10.bin",
        "vgj5ic13.bin",
        "sk4_ic14.bin",
        "sk4_ic15.bin",
        "sk4_ic16.bin"
    ]
    func_map['maincpu'] = helpers.equal_split_helper('vanguardj.maincpu', maincpu_filenames)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("vanguardj.zip", bundle_contents, func_map))

    return out_files

def _handle_pow(bundle_contents):
    # POW Common
    func_map = {}
    out_files = []

    gfx1_filenames = [
        "dg9.l25",
        "dg10.m25"
    ]
    func_map['gfx1'] = helpers.equal_split_helper('pow.gfx1', gfx1_filenames)

    func_map['soundcpu'] = helpers.name_file_helper('pow.soundcpu', 'dg8.e25')
    func_map['upd'] = helpers.name_file_helper('pow.upd', 'dg7.d20')
    func_map['plds'] = helpers.name_file_helper('pow.plds', 'pal20l10.a6')

    gfx2_file_names = [
        "snk88011a.1a",
        "snk88015a.2a",
        "snk88012a.1b",
        "snk88016a.2b",
        "snk88013a.1c",
        "snk88017a.2c",
        "snk88014a.1d",
        "snk88018a.2d",
        "snk88019a.3a",
        "snk88023a.4a",
        "snk88020a.3b",
        "snk88024a.4b",
        "snk88021a.3c",
        "snk88025a.4c",
        "snk88022a.3d",
        "snk88026a.4d"
    ]
    def pow_gfx2_common(in_file_name, filenames):
        def gfx2(in_files):
            contents = in_files[in_file_name]
            chunks = transforms.equal_split(contents, 8)
            chunks = transforms.deinterleave_all(chunks, num_ways=2, word_size=1)
            return dict(zip(filenames, chunks))
        return gfx2
    func_map['gfx2'] = pow_gfx2_common('pow.gfx2', gfx2_file_names)
    logger.info("Processing POW common files...")
    common_file_map = helpers.process_rom_files(bundle_contents, func_map)

    # POW
    func_map = {}
    maincpu_filenames = [
        "dg1ver1.j14",
        "dg2ver1.l14"
    ]
    func_map['maincpu'] = helpers.deinterleave_helper('pow.maincpu', maincpu_filenames, 2, 1)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("pow.zip", bundle_contents, func_map))

    # POWJ
    func_map = {}
    maincpu_filenames = [
        "1-2",
        "2-2"
    ]
    func_map['maincpu'] = helpers.deinterleave_helper('powj.maincpu', maincpu_filenames, 2, 1)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("powj.zip", bundle_contents, func_map))

    return out_files

def _handle_victoryroad(bundle_contents, patch_contents):
    '''Extract Ikari 2 Victory Road'''
    out_files = []

    # VICTROAD Common
    func_map = {}
    func_map['sub'] = helpers.name_file_helper("VictoryRoad.1.z80", "p2.8p")
    sp32_filenames = [
        "p11.4m",
        "p14.2m",
        "p12.4p",
        "p15.2p",
        "p13.4r",
        "p16.2r"
    ]
    func_map['sp32'] = helpers.equal_split_helper('VictoryRoad.sp32', sp32_filenames)
    bg_filenames = [
        "p17.4c",
        "p18.2c",
        "p19.4b",
        "p20.2b"
    ]
    func_map['bg'] = helpers.equal_split_helper('VictoryRoad.bg', bg_filenames)
    logger.info("Processing Victory Road common files...")
    common_file_map = helpers.process_rom_files(bundle_contents, func_map)

    # VICTROAD PID Special copy from CHOPPERB
    # Missing victroad:a5004-1.6d is chopperb:p-a1.2c
    func_map = {}
    func_map['plds'] = helpers.name_file_helper("chopper.plds", "a5004-1.6d")
    logger.info("Processing Victory Road common files...")
    victroad_pid_file_map = helpers.process_rom_files(patch_contents, func_map)

    # DOGOSOKE
    func_map = {}
    func_map['maincpu'] = helpers.name_file_helper("VictoryRoad.j.0.z80", "p1.4p")
    func_map['audiocpu'] = helpers.name_file_helper("VictoryRoad.j.2.z80", "p3.7k")
    adpcm_filenames = [
        "p4.5e",
        "p5.5g"
    ]
    func_map['adpcm'] = helpers.equal_split_helper('VictoryRoad.j.adpcm', adpcm_filenames)
    func_map['tx'] = helpers.name_file_helper("VictoryRoad.j.tx", "p7.3b")
    sp_filenames = [
        "p8.3d",
        "p9.3f",
        "p10.3h"
    ]
    func_map['sp'] = helpers.equal_split_helper('VictoryRoad.j.sp', sp_filenames)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    ph_files = {
        "1.1d": 4096,
        "1.2d": 4096
    }
    func_map['placeholders'] = helpers.placeholder_helper(ph_files)
    pal_filenames = [
        "c2.2l",
        "c1.1k",
        "c3.1l"
    ]
    func_map['pal'] = utils.simple_palette_helper('VictoryRoad.j.pal', pal_filenames)
    out_files.append(utils.build_snk_rom("dogosoke.zip", bundle_contents, func_map))

    # VICTROAD
    func_map = {}
    func_map['pid'] = helpers.existing_files_helper(victroad_pid_file_map)
    func_map['maincpu'] = helpers.name_file_helper("VictoryRoad.0.z80", "p1.4p")
    func_map['audiocpu'] = helpers.name_file_helper("VictoryRoad.2.z80", "p3.7k")
    adpcm_filenames = [
        "p4.5e",
        "p5.5g"
    ]
    func_map['adpcm'] = helpers.equal_split_helper('VictoryRoad.adpcm', adpcm_filenames)
    func_map['tx'] = helpers.name_file_helper("VictoryRoad.tx", "p7.3b")
    sp_filenames = [
        "p8.3d",
        "p9.3f",
        "p10.3h"
    ]
    func_map['sp'] = helpers.equal_split_helper('VictoryRoad.sp', sp_filenames)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    ph_files = {
        "1.1d": 4096,
        "1.2d": 4096,
        "a5004-4.8s": 260,
        "a6002-2.5r": 324,
        "a6002-3.2p": 260
    }
    func_map['placeholders'] = helpers.placeholder_helper(ph_files)
    pal_filenames = [
        "c2.2l",
        "c1.1k",
        "c3.1l"
    ]
    func_map['pal'] = utils.simple_palette_helper('VictoryRoad.pal', pal_filenames)
    out_files.append(utils.build_snk_rom("victroad.zip", bundle_contents, func_map))

    return out_files

def _handle_ikari(bundle_contents):
    '''Extract Ikari Warriors'''
    out_files = []

    # Ikari Common
    func_map = {}
    sp32_filenames = [
        "p11.4m",
        "p14.2m",
        "p12.4p",
        "p15.2p",
        "p13.4r",
        "p16.2r"
    ]
    func_map['sp32'] = helpers.equal_split_helper('IkariWarriors.sp32', sp32_filenames)
    logger.info("Processing Ikari common files...")
    common_file_map = helpers.process_rom_files(bundle_contents, func_map)

    # Ikari Palette Common
    func_map = {}
    pal_filenames = [
        "2.2j",
        "1.1h",
        "3.1j"
    ]
    func_map['pal'] = utils.simple_palette_helper('IkariWarriors.pal', pal_filenames)
    logger.info("Processing Ikari Palette common files...")
    palette_common_file_map = helpers.process_rom_files(bundle_contents, func_map)

    # Ikari US Common
    func_map = {}
    func_map['tx'] = helpers.name_file_helper("IkariWarriors.tx", "p7.3b")
    sp_filenames = [
        "p8.3d",
        "p9.3f",
        "p10.3h"
    ]
    func_map['sp'] = helpers.equal_split_helper('IkariWarriors.sp', sp_filenames)
    bg_filenames = [
        "p17.4d",
        "p18.2d",
        "p19.4b",
        "p20.2b"
    ]
    func_map['bg'] = helpers.equal_split_helper('IkariWarriors.bg', bg_filenames)
    logger.info("Processing Ikari US common files...")
    us_common_file_map = helpers.process_rom_files(bundle_contents, func_map)

    # Ikari Sub Common
    func_map = {}
    sub_file_map = {
        "p3.8l": 0x4000,
        "p4.8k": 0x8000
    }
    func_map['audiocpu'] = helpers.custom_split_helper('IkariWarriors.a.1.z80', sub_file_map)
    logger.info("Processing Ikari Sub common files...")
    sub_common_file_map = helpers.process_rom_files(bundle_contents, func_map)

    # Ikari AudioCPU Common
    func_map = {}
    audiocpu_file_map = {
        "p5.6e": 0x4000,
        "p6.6f": 0x8000
    }
    func_map['audiocpu'] = helpers.custom_split_helper('IkariWarriors.a.2.z80', audiocpu_file_map)
    logger.info("Processing Ikari AudioCPU Common files...")
    audiocpu_common_file_map = helpers.process_rom_files(bundle_contents, func_map)

    # IKARIA
    func_map = {}
    maincpu_file_map = {
        "p1.4l": 0x4000,
        "p2.4k": 0x8000
    }
    func_map['maincpu'] = helpers.custom_split_helper('IkariWarriors.a.0.z80', maincpu_file_map)
    func_map['sub'] = helpers.existing_files_helper(sub_common_file_map)
    func_map['audiocpu'] = helpers.existing_files_helper(audiocpu_common_file_map)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    func_map['us_common'] = helpers.existing_files_helper(us_common_file_map)
    func_map['palette'] = helpers.existing_files_helper(palette_common_file_map)
    ph_files = {
        "ampal16l8a-a5004-3.2n": 260,
        "ampal16l8a-a5004-4.8s": 260,
        "pal20l8a-a5004-2.6m": 324,
        "ampal16r6a-a5004-1.6d": 260
    }
    func_map['placeholders'] = helpers.placeholder_helper(ph_files)
    out_files.append(utils.build_snk_rom("ikaria.zip", bundle_contents, func_map))

    # IKARIJP
    func_map = {}
    maincpu_file_map = {
        "p1.4l": 0x4000,
        "p2.4k": 0x8000
    }
    func_map['maincpu'] = helpers.custom_split_helper('IkariWarriors.j.0.z80', maincpu_file_map)
    sub_file_map = {
        "p3.8l": 0x4000,
        "p4.8k": 0x8000
    }
    func_map['sub'] = helpers.custom_split_helper('IkariWarriors.j.1.z80', sub_file_map)
    func_map['audiocpu'] = helpers.existing_files_helper(audiocpu_common_file_map)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    func_map['tx'] = helpers.name_file_helper("IkariWarriors.j.tx", "p7.3b")
    sp_filenames = [
        "p8.3d",
        "p9.3f",
        "p10.3h"
    ]
    func_map['sp'] = helpers.equal_split_helper('IkariWarriors.j.sp', sp_filenames)
    bg_filenames = [
        "p17.4d",
        "p18.2d",
        "p19.4b",
        "p20.2b"
    ]
    func_map['bg'] = helpers.equal_split_helper('IkariWarriors.j.bg', bg_filenames)
    func_map['palette'] = helpers.existing_files_helper(palette_common_file_map)
    ph_files = {
        "ampal16l8a-a5004-3.2n": 260,
        "ampal16l8a-a5004-4.8s": 260,
        "pal20l8a-a5004-2.6m": 324,
        "ampal16r6a-a5004-1.6d": 260
    }
    func_map['placeholders'] = helpers.placeholder_helper(ph_files)
    out_files.append(utils.build_snk_rom("ikarijp.zip", bundle_contents, func_map))

    # IKARI
    func_map = {}
    func_map['maincpu'] = helpers.name_file_helper("IkariWarriors.0.z80", "1.4p")
    func_map['sub'] = helpers.name_file_helper("IkariWarriors.1.z80", "2.8p")
    func_map['audiocpu'] = helpers.name_file_helper("IkariWarriors.2.z80", "3.7k")
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    func_map['us_common'] = helpers.existing_files_helper(us_common_file_map)
    func_map['palette'] = helpers.common_rename_helper(palette_common_file_map, {
        "2.2j": "a6002-2.2l",
        "1.1h": "a6002-1.1k",
        "3.1j": "a6002-3.1l"
    })
    out_files.append(utils.build_snk_rom("ikari.zip", bundle_contents, func_map))

    # IKARINC
    func_map = {}
    maincpu_file_map = {
        "p1.4l": 0x4000,
        "p2.4k": 0x8000
    }
    func_map['maincpu'] = helpers.custom_split_helper('IkariWarriors.nc.0.z80', maincpu_file_map)
    func_map['sub'] = helpers.existing_files_helper(sub_common_file_map)
    func_map['audiocpu'] = helpers.existing_files_helper(audiocpu_common_file_map)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    func_map['us_common'] = helpers.existing_files_helper(us_common_file_map)
    func_map['palette'] = helpers.existing_files_helper(palette_common_file_map)
    ph_files = {
        "ampal16l8a-a5004-3.2n": 260,
        "ampal16l8a-a5004-4.8s": 260,
        "pal20l8a-a5004-2.6m": 324,
        "ampal16r6a-a5004-1.6d": 260
    }
    func_map['placeholders'] = helpers.placeholder_helper(ph_files)
    out_files.append(utils.build_snk_rom("ikarinc.zip", bundle_contents, func_map))

    return out_files

def _handle_guerilla(bundle_contents):
    '''Extract Guerilla Wars'''
    out_files = []

    # GWAR Common
    func_map = {}
    sp_filenames = [
        "gw6.2j",
        "7.2l",
        "gw8.2m",
        "gw9.2p"
    ]
    func_map['sp'] = helpers.equal_split_helper('GuerillaWar.sp', sp_filenames)
    sp32_filenames = [
        "16.2ab",
        "17.2ad",
        "14.2y",
        "15.2aa",
        "12.2v",
        "13.2w",
        "10.2s",
        "11.2t"
    ]
    func_map['sp32'] = helpers.equal_split_helper('GuerillaWar.sp32', sp32_filenames)
    bg_filenames = [
        "18.8x",
        "19.8z",
        "gw20.8aa",
        "21.8ac"
    ]
    func_map['bg'] = helpers.equal_split_helper('GuerillaWar.bg', bg_filenames)
    func_map['adpcm'] = helpers.name_file_helper("GuerillaWar.adpcm", "4.2j")
    logger.info("Processing GWAR common files...")
    common_file_map = helpers.process_rom_files(bundle_contents, func_map)

    # GWAR Palette Common
    func_map = {}
    pal_filenames = [
        '2.9v',
        '3.9w',
        '1.9u'
    ]
    func_map['pal'] = utils.simple_palette_helper('GuerillaWar.pal', pal_filenames)
    logger.info("Processing GWAR palette common files...")
    pal_common_file_map = helpers.process_rom_files(bundle_contents, func_map)

    # GWAR Audio Common
    func_map = {}
    func_map['sub'] = helpers.name_file_helper("GuerillaWar.1.z80", "2.6g")
    func_map['audiocpu'] = helpers.name_file_helper("GuerillaWar.2.z80", "3.7g")
    logger.info("Processing GWAR audio common files...")
    audio_common_file_map = helpers.process_rom_files(bundle_contents, func_map)

    # GWAR
    func_map = {}
    func_map['tx'] = helpers.name_file_helper("GuerillaWar.tx", "gw5.8p")
    func_map['maincpu'] = helpers.name_file_helper("GuerillaWar.0.z80", "1.2g")
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    func_map['audio'] = helpers.existing_files_helper(audio_common_file_map)
    func_map['pal'] = helpers.existing_files_helper(pal_common_file_map)
    ph_files = {
        'l.1x': 0x1000,
        'l.1w': 0x1000
    }
    func_map['placeholders'] = helpers.placeholder_helper(ph_files)
    out_files.append(utils.build_snk_rom("gwar.zip", bundle_contents, func_map))

    # # GWARA
    func_map = {}
    func_map['tx'] = helpers.name_file_helper("GuerillaWar.tx", "gv5.3a")
    func_map['maincpu'] = helpers.name_file_helper("GuerillaWar.a.0.z80", "gv3_1.4p")
    func_map['sub'] = helpers.name_file_helper("GuerillaWar.a.1.z80", "gv4.8p")
    func_map['audiocpu'] = helpers.name_file_helper("GuerillaWar.a.2.z80", "gv2.7k")
    func_map['common'] = helpers.common_rename_helper(common_file_map, {
        "4.2j": "gv1.5g",
        "gw6.2j": "gv9.3g",
        "7.2l": "gv8.3e",
        "gw8.2m": "gv7.3d",
        "gw9.2p": "gv6.3b",
        "16.2ab": "gv14.8l",
        "17.2ad": "gv15.8n",
        "14.2y": "gv16.8p",
        "15.2aa": "gv17.8s",
        "12.2v": "gv18.7p",
        "13.2w": "gv19.7s",
        "10.2s": "gv20.8j",
        "11.2t": "gv21.8k",
        "18.8x": "gv13.2a",
        "19.8z": "gv12.2b",
        "gw20.8aa": "gv11.2d",
        "21.8ac": "gv10.2e"
    })
    func_map['pal'] = helpers.common_rename_helper(pal_common_file_map, {
        "3.9w": "1.1k",
        "1.9u": "2.1l",
        "2.9v": "3.2l"
    })
    ph_files = {
        'l.1x': 0x1000,
        'l.1w': 0x1000,
        'horizon.8j': 0x400,
        'vertical.8k': 0x400
    }
    func_map['placeholders'] = helpers.placeholder_helper(ph_files)
    out_files.append(utils.build_snk_rom("gwara.zip", bundle_contents, func_map))

    # GWARAB
    func_map = {}
    func_map['tx'] = helpers.name_file_helper("GuerillaWar.tx", "gv5.3a")
    func_map['maincpu'] = helpers.name_file_helper("GuerillaWar.b.0.z80", "gv3 ver 1.4p")
    func_map['sub'] = helpers.name_file_helper("GuerillaWar.a.1.z80", "gv4.8p")
    func_map['audiocpu'] = helpers.name_file_helper("GuerillaWar.a.2.z80", "gv2.7k")
    func_map['common'] = helpers.common_rename_helper(common_file_map, {
        "4.2j": "gv1.5g",
        "gw6.2j": "gv9.3g",
        "7.2l": "gv8.3e",
        "gw8.2m": "gv7.3d",
        "gw9.2p": "gv6.3b",
        "16.2ab": "gv14.8l",
        "17.2ad": "gv15.8n",
        "14.2y": "gv16.8p",
        "15.2aa": "gv17.8s",
        "12.2v": "gv18.7p",
        "13.2w": "gv19.7s",
        "10.2s": "gv20.8j",
        "11.2t": "gv21.8k",
        "18.8x": "gv13.2a",
        "19.8z": "gv12.2b",
        "gw20.8aa": "gv11.2d",
        "21.8ac": "gv10.2e"
    })
    func_map['pal'] = helpers.common_rename_helper(pal_common_file_map, {
        "3.9w": "1.1k",
        "1.9u": "2.1l",
        "2.9v": "3.2l"
    })
    ph_files = {
        'l.1x': 0x1000,
        'l.1w': 0x1000,
        'horizon.8j': 0x400,
        'vertical.8k': 0x400
    }
    func_map['placeholders'] = helpers.placeholder_helper(ph_files)
    out_files.append(utils.build_snk_rom("gwarab.zip", bundle_contents, func_map))

    # GWARJ
    func_map = {}
    func_map['tx'] = helpers.name_file_helper("GuerillaWar.j.tx", "gw5.8p")
    func_map['maincpu'] = helpers.name_file_helper("GuerillaWar.j.0.z80", "1.2g")
    func_map['audio'] = helpers.existing_files_helper(audio_common_file_map)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    func_map['pal'] = helpers.existing_files_helper(pal_common_file_map)
    ph_files = {
        'l.1x': 0x1000,
        'l.1w': 0x1000
    }
    func_map['placeholders'] = helpers.placeholder_helper(ph_files)
    out_files.append(utils.build_snk_rom("gwarj.zip", bundle_contents, func_map))

    return out_files

def _handle_psycho(bundle_contents):
    '''Extract Psycho Soldier'''
    out_files = []

    # PSYCHO Common
    func_map = {}
    func_map['sub'] = helpers.name_file_helper("PsychoSoldier.1.z80", "ps6.8m")
    func_map['tx'] = helpers.name_file_helper("PsychoSoldier.tx", "ps8.3a")
    bg_filenames = [
        "ps16.1f",
        "ps15.1d",
        "ps14.1c",
        "ps13.1a"
    ]
    func_map['bg'] = helpers.equal_split_helper('PsychoSoldier.bg', bg_filenames)
    sp_filenames = [
        "ps12.3g",
        "ps11.3e",
        "ps10.3c",
        "ps9.3b"
    ]
    func_map['sp'] = helpers.equal_split_helper('PsychoSoldier.sp', sp_filenames)
    sp32_filenames = [
        "ps17.10f",
        "ps18.10h",
        "ps19.10j",
        "ps20.10l",
        "ps21.10m",
        "ps22.10n",
        "ps23.10r",
        "ps24.10s"
    ]
    func_map['sp32'] = helpers.equal_split_helper('PsychoSoldier.sp32', sp32_filenames)
    ph_files = {
        'horizon.8j': 0x400,
        'vertical.8k': 0x400
    }
    func_map['ph'] = helpers.placeholder_helper(ph_files)
    pal_filenames = [
        "psc3.1l",
        "psc1.1k",
        "psc2.2k"
    ]
    func_map['pal'] = utils.simple_palette_helper('PsychoSoldier.pal', pal_filenames)
    logger.info("Processing Psycho Soldier common files...")
    common_file_map = helpers.process_rom_files(bundle_contents, func_map)

    # PSYCHOS
    func_map = {}
    func_map['maincpu'] = helpers.name_file_helper("PsychoSoldier.0.z80", "ps7.4m")
    func_map['sub'] = helpers.name_file_helper("PsychoSoldier.2.z80", "ps5.6j")
    adpcm_filenames = [
        "ps1.5b",
        "ps2.5c",
        "ps3.5d",
        "ps4.5f"
    ]
    func_map['adpcm'] = helpers.equal_split_helper('PsychoSoldier.adpcm', adpcm_filenames)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("psychos.zip", bundle_contents, func_map))

    # PSYCHOSJ
    func_map = {}
    func_map['maincpu'] = helpers.name_file_helper("PsychoSoldier.j.0.z80", "ps7.4m")
    func_map['sub'] = helpers.name_file_helper("PsychoSoldier.j.2.z80", "ps5.6j")
    adpcm_filenames = [
        "ps1.5b",
        "ps2.5c",
        "ps3.5d",
        "ps4.5f"
    ]
    func_map['adpcm'] = helpers.equal_split_helper('PsychoSoldier.j.adpcm', adpcm_filenames)
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("psychosj.zip", bundle_contents, func_map))

    return out_files

def _handle_aso(bundle_contents):
    '''Extract Armored Scrum Object'''
    out_files = []
    for key, value in bundle_contents.items():
        if key.startswith("ASOArmoredScrumObject"):
            print(f'{key}: {len(value)}')

    # ASO Common
    func_map = {}
    func_map['bg'] = helpers.name_file_helper("ASOArmoredScrumObject.bg", "p10.14h")
    audiocpu_filenames = [
        "p7.4f",
        "p8.3f",
        "p9.2f"
    ]
    func_map['audiocpu'] = helpers.equal_split_helper(
        'ASOArmoredScrumObject.2.z80', audiocpu_filenames)
    sp_filenames = [
        "p11.11h",
        "p12.9h",
        "p13.8h"
    ]
    def aso_sp(in_files):
        contents = in_files['ASOArmoredScrumObject.sp']
        chunks = transforms.equal_split(contents, num_chunks = 12)

        p11 = transforms.merge([chunks[2] + chunks[3] + chunks[0] + chunks[1]])
        p12 = transforms.merge([chunks[6] + chunks[7] + chunks[4] + chunks[5]])
        p13 = transforms.merge([chunks[10] + chunks[11] + chunks[8] + chunks[9]])

        chunks = [p11, p12, p13]

        return dict(zip(sp_filenames, chunks))
    func_map['sp'] = aso_sp
    pals_filenames = [
        "mb7122h.13f",
        "mb7122h.12f",
        "mb7122h.14f"
    ]
    func_map['pal'] = utils.palette_rebuild_helper(pals_filenames, 'ASOArmoredScrumObject.pal')
    ph_files = {
        'pal16l8a-1.bin': 260,
        'pal16l8a-2.bin': 260,
        'pal16r6a.15b': 260
    }
    func_map['ph'] = helpers.placeholder_helper(ph_files)
    logger.info("Processing ASO common files...")
    common_file_map = helpers.process_rom_files(bundle_contents, func_map)

    # ASO
    func_map = {}
    maincpu_filenames = [
        "p1.8d",
        "p2.7d",
        "p3.5d"
    ]
    func_map['maincpu'] = helpers.equal_split_helper(
        'ASOArmoredScrumObject.0.z80', maincpu_filenames)
    sub_filenames = [
        "p4.3d",
        "p5.2d",
        "p6.1d"
    ]
    func_map['sub'] = helpers.equal_split_helper('ASOArmoredScrumObject.1.z80', sub_filenames)
    func_map['tx'] = helpers.name_file_helper("ASOArmoredScrumObject.tx", "p14.1h")
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("aso.zip", bundle_contents, func_map))

    # ALPHAMIS
    func_map = {}
    maincpu_filenames = [
        "p1.8d",
        "p2.7d",
        "p3.5d"
    ]
    func_map['maincpu'] = helpers.equal_split_helper(
        'ASOArmoredScrumObject.b.0.z80', maincpu_filenames)
    sub_filenames = [
        "p4.3d",
        "p5.2d",
        "p6.1d"
    ]
    func_map['sub'] = helpers.equal_split_helper('ASOArmoredScrumObject.b.1.z80', sub_filenames)
    func_map['tx'] = helpers.name_file_helper("ASOArmoredScrumObject.b.tx", "p14.1h")
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("alphamis.zip", bundle_contents, func_map))

    # ARIAN
    func_map = {}
    maincpu_filenames = [
        "p1.8d",
        "p2.7d",
        "p3.5d"
    ]
    func_map['maincpu'] = helpers.equal_split_helper(
        'ASOArmoredScrumObject.c.0.z80', maincpu_filenames)
    sub_filenames = [
        "p4.3d",
        "p5.2d",
        "p6.1d"
    ]
    func_map['sub'] = helpers.equal_split_helper('ASOArmoredScrumObject.c.1.z80', sub_filenames)
    func_map['tx'] = helpers.name_file_helper("ASOArmoredScrumObject.c.tx", "p14.1h")
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("arian.zip", bundle_contents, func_map))

    return out_files

def _handle_athena(bundle_contents):
    '''Extract Athena'''
    out_files = []

    func_map = {}
    sub_file_map = {
        "p3.8p": 0x4000,
        "p4.8m": 0x8000
    }
    func_map['sub'] = helpers.custom_split_helper('Athena.1.z80', sub_file_map)
    audiocpu_file_map = {
        "p5.6g": 0x4000,
        "p6.6k": 0x8000
    }
    func_map['audiocpu'] = helpers.custom_split_helper('Athena.2.z80', audiocpu_file_map)
    sp_filenames = [
        "p7.2p",
        "p8.2s",
        "p9.2t"
    ]
    func_map['sp'] = helpers.equal_split_helper('Athena.sp', sp_filenames)

    maincpu_file_map = {
        "p1.4p": 0x4000,
        "p2.4m": 0x8000
    }
    func_map['maincpu'] = helpers.custom_split_helper('Athena.0.z80', maincpu_file_map)
    func_map['tx'] = helpers.name_file_helper("Athena.tx", "p11.2d")
    func_map['bg'] = helpers.name_file_helper("Athena.bg", "p10.2b")
    pals_filenames = [
        "2.1b",
        "3.2c",
        "1.1c"
    ]
    func_map['pal'] = utils.palette_rebuild_helper(pals_filenames, 'Athena.pal')
    out_files.append(utils.build_snk_rom("athena.zip", bundle_contents, func_map))

    return out_files

def _handle_tnk3(bundle_contents):
    '''Extract TnkIII'''
    out_files = []

    # # TNK3 Common
    func_map = {}
    sub_filenames = [
        "p4.2e",
        "p5.2f",
        "p6.2h"
    ]
    func_map['sub'] = helpers.equal_split_helper('TNKIII.1.z80', sub_filenames)
    audiocpu_filenames = [
        "p10.6f",
        "p11.6d"
    ]
    func_map['audiocpu'] = helpers.equal_split_helper('TNKIII.2.z80', audiocpu_filenames)
    bg_filenames = [
        "p12.3d",
        "p13.3c"
    ]
    func_map['bg'] = helpers.equal_split_helper('TNKIII.bg', bg_filenames)
    sp_filenames = [
        "p7.7h",
        "p8.7f",
        "p9.7e"
    ]
    func_map['sp'] = helpers.equal_split_helper('TNKIII.sp', sp_filenames)
    pals_filenames = [
        "1.5g",
        "2.5f",
        "0.5h"
    ]
    func_map['pal'] = utils.palette_rebuild_helper(pals_filenames, 'TNKIII.pal')
    logger.info("Processing TNK3 common files...")
    common_file_map = helpers.process_rom_files(bundle_contents, func_map)

    # # TNK3
    func_map = {}
    maincpu_filenames = [
        "p1.4e",
        "p2.4f",
        "p3.4h"
    ]
    func_map['maincpu'] = helpers.equal_split_helper('TNKIII.0.z80', maincpu_filenames)
    func_map['tx'] = helpers.name_file_helper("TNKIII.tx", "p14.1e")
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("tnk3.zip", bundle_contents, func_map))

    # TNK3J
    func_map = {}
    maincpu_filenames = [
        "p1.4e",
        "p2.4f",
        "p3.4h"
    ]
    func_map['maincpu'] = helpers.equal_split_helper('TNKIII.j.0.z80', maincpu_filenames)
    func_map['tx'] = helpers.name_file_helper("TNKIII.j.tx", "p14.1e")
    func_map['common'] = helpers.existing_files_helper(common_file_map)
    out_files.append(utils.build_snk_rom("tnk3j.zip", bundle_contents, func_map))

    return out_files



import logging
import click
import click_log

from gex.lib.utils.blob import transforms

logger = logging.getLogger('gextoolbox')

@click.command()
@click.option('--in', 'in_file', help = 'path to input file', required=True)
@click.option('--out', 'out_file_base',
    help = 'path to output file base name (ex. ./my.file)', required=True)
@click.option('--ways', 'num_ways',
    help = 'number of ways to deinterleave, default is 2-way', default=2)
@click.option('--word', 'word_size',
    help = 'size of word (number of bytes) to put in each file, default is 2 bytes', default=2)
@click_log.simple_verbosity_option(logger)
def deinterleave(in_file, out_file_base, num_ways, word_size):
    '''CLI tool to deinterleave a file into num_ways files by word_size bytes'''
    in_data = read_bin_file(in_file)
    deinterleaved_chunks = transforms.deinterleave(in_data, num_ways, word_size)

    for idx, chunk in enumerate(deinterleaved_chunks):
        filename = f'{out_file_base}.{idx}'
        write_bin_file(chunk, filename)

    logger.info("Done.")

def read_bin_file(path):
    try:
        with open(path, "rb") as file:
            content = file.read()
            return content
    except IOError:
        logger.error(f"Error reading {path}!")
        exit()

def write_bin_file(data, path):
    try:
        with open(path, "wb") as file:
            file.write(data)
    except IOError:
        logger.error(f"Error writing {path}!")
        exit()



'''
Reader for Apple Binary PList

### shawngmc NOTE ###
Pulled from https://github.com/ValadAmoleo/sf30ac-extractor/blob/mame/bplist.py
Altered to:
* Be much more pythonic/linted/etc.

#### WydD NOTE ####
Original project https://github.com/farcaller/bplist-python
This has been altered to
* Be Py3k compatible
* Support the specific tail format of the sf30ac files

#################################################################################
# Copyright (C) 2009-2011 Vladimir "Farcaller" Pouzanov <farcaller@gmail.com>   #
#                                                                               #
# Permission is hereby granted, free of charge, to any person obtaining a copy  #
# of this software and associated documentation files (the "Software"), to deal #
# in the Software without restriction, including without limitation the rights  #
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell     #
# copies of the Software, and to permit persons to whom the Software is         #
# furnished to do so, subject to the following conditions:                      #
#                                                                               #
# The above copyright notice and this permission notice shall be included in    #
# all copies or substantial portions of the Software.                           #
#                                                                               #
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR    #
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,      #
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE   #
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER        #
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, #
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN     #
# THE SOFTWARE.                                                                 #
#################################################################################
'''

import struct
from datetime import datetime, timedelta
import plistlib

class BPListReader(object):
    '''Reader for Apple Binary PList'''
    def __init__(self, in_data):
        self.data = in_data
        self.objects = []
        self.resolved = {}
        self.offset_size = None
        self.object_ref_size = None
        self.number_of_objects = None
        self.top_object = None
        self.table_offset = None
        self.top_object = None
        self.offset_table = None
        self.offsets = None

    def __unpack_int_struct(self, size, in_data):
        '''__unpack_int_struct(size, string) -> int

        Unpacks the integer of given size (1, 2 or 4 bytes) from string
        '''
        if size == 1:
            int_format = '!B'
        elif size == 2:
            int_format = '!H'
        elif size == 4:
            int_format = '!I'
        elif size == 8:
            int_format = '!Q'
        else:
            raise Exception('int unpack size '+str(size)+' unsupported')
        return struct.unpack(int_format, in_data)[0]

    def __unpack_int(self, offset):
        '''__unpack_int(offset) -> int

        Unpacks int field from plist at given offset
        '''
        return self.__unpack_int_meta(offset)[1]

    def __unpack_int_meta(self, offset):
        '''__unpack_int_meta(offset) -> (size, int)

        Unpacks int field from plist at given offset and returns its size and value
        '''
        obj_header = struct.unpack('!B', self.data[offset:offset+1])[0]
        _, obj_info = (obj_header & 0xF0), (obj_header & 0x0F)
        int_sz = 2**obj_info
        return int_sz, self.__unpack_int_struct(int_sz, self.data[offset+1:offset+1+int_sz])

    def __resolve_int_size(self, obj_info, offset):
        '''__resolve_int_size(obj_info, offset) -> (count, offset)

        Calculates count of objref* array entries and returns count and offset to first element
        '''
        if obj_info == 0x0F:
            ofs, obj_count = self.__unpack_int_meta(offset+1)
            objref = offset+2+ofs
        else:
            obj_count = obj_info
            objref = offset+1
        return obj_count, objref

    def __unpack_float_struct(self, size, in_data):
        '''__unpack_float_struct(size, string) -> float

        Unpacks the float of given size (4 or 8 bytes) from string
        '''
        if size == 4:
            float_format = '!f'
        elif size == 8:
            float_format = '!d'
        else:
            raise Exception('float unpack size '+str(size)+' unsupported')
        return struct.unpack(float_format, in_data)[0]

    def __unpack_float(self, offset):
        '''__unpack_float(offset) -> float

        Unpacks float field from plist at given offset
        '''
        obj_header = struct.unpack('!B', self.data[offset])[0]
        _, obj_info = (obj_header & 0xF0), (obj_header & 0x0F)
        int_sz = 2**obj_info
        return int_sz, self.__unpack_float_struct(int_sz, self.data[offset+1:offset+1+int_sz])

    def __unpack_date(self, offset):
        stored_time_delta = int(struct.unpack(">d", self.data[offset+1:offset+9])[0])
        return datetime(year=2001,month=1,day=1) + timedelta(seconds=stored_time_delta)

    def __unpack_item(self, offset):
        '''__unpack_item(offset)

        Unpacks and returns an item from plist
        '''
        obj_header = struct.unpack('!B', self.data[offset:offset+1])[0]
        obj_type, obj_info = (obj_header & 0xF0), (obj_header & 0x0F)
        if   obj_type == 0x00:
            if   obj_info == 0x00:
                # null   0000 0000
                return None
            elif obj_info == 0x08:
                # bool   0000 1000           // false
                return False
            elif obj_info == 0x09:
                # bool   0000 1001           // true
                return True
            elif obj_info == 0x0F:
                # fill   0000 1111           // fill byte
                raise Exception("0x0F Not Implemented") # this is really pad byte, FIXME
            else:
                raise Exception('unpack item type '+str(obj_header)+' at '+str(offset)+ 'failed')
        elif obj_type == 0x10:
            #     int    0001 nnnn   ...
            # # of bytes is 2^nnnn, big-endian bytes
            return self.__unpack_int(offset)
        elif obj_type == 0x20:
            #    real    0010 nnnn   ...
            # # of bytes is 2^nnnn, big-endian bytes
            return self.__unpack_float(offset)
        elif obj_type == 0x30:
            #    date    0011 0011   ...
            # 8 byte float follows, big-endian bytes
            return self.__unpack_date(offset)
        elif obj_type == 0x40:
            #    data    0100 nnnn   [int]   ...
            # nnnn is number of bytes unless 1111 then int count follows, followed by bytes
            obj_count, objref = self.__resolve_int_size(obj_info, offset)
            return self.data[objref:objref+obj_count] # we return data as str
        elif obj_type == 0x50:
            #    string  0101 nnnn   [int]   ...
            # ASCII string, nnnn is # of chars, else 1111
            # then int count, then bytes
            obj_count, objref = self.__resolve_int_size(obj_info, offset)
            return self.data[objref:objref+obj_count].decode('ascii')
        elif obj_type == 0x60:
            #    string  0110 nnnn   [int]   ...
            # Unicode string, nnnn is # of chars, else 1111
            # then int count, then big-endian 2-byte uint16_t
            obj_count, objref = self.__resolve_int_size(obj_info, offset)
            return self.data[objref:objref+obj_count*2].decode('utf-16be')
        elif obj_type == 0x80:
            #    uid     1000 nnnn   ...     // nnnn+1 is # of bytes
            # FIXME: Accept as a string for now
            obj_count, objref = self.__resolve_int_size(obj_info, offset)
            return self.data[objref:objref+obj_count]
        elif obj_type == 0xA0:
            #    array   1010 nnnn   [int]   objref*
            # nnnn is count, unless '1111', then int count follows
            obj_count, objref = self.__resolve_int_size(obj_info, offset)
            arr = []
            for i in range(obj_count):
                arr.append(
                    self.__unpack_int_struct(
                        self.object_ref_size,
                        self.data[objref + i * self.object_ref_size:
                            objref + i * self.object_ref_size + self.object_ref_size]
                    )
                )
            return arr
        elif obj_type == 0xC0:
            #   set      1100 nnnn   [int]   objref*
            # nnnn is count, unless '1111', then int count follows
            # not serializable via apple implementation
            raise Exception("0xC0 Not Implemented") # FIXME: implement
        elif obj_type == 0xD0:
            #   dict     1101 nnnn   [int]   keyref* objref*
            # nnnn is count, unless '1111', then int count follows
            obj_count, objref = self.__resolve_int_size(obj_info, offset)
            keys = []
            for i in range(obj_count):
                keys.append(
                    self.__unpack_int_struct(
                        self.object_ref_size,
                        self.data[objref + i * self.object_ref_size:
                            objref + i * self.object_ref_size +self.object_ref_size]
                    )
                )
            values = []
            objref += obj_count*self.object_ref_size
            for i in range(obj_count):
                values.append(
                    self.__unpack_int_struct(
                        self.object_ref_size,
                        self.data[objref + i * self.object_ref_size:
                            objref + i * self.object_ref_size + self.object_ref_size]
                    )
                )
            dic = {}
            for i in range(obj_count):
                dic[keys[i]] = values[i]
            return dic
        else:
            raise Exception('don\'t know how to unpack obj type '+hex(obj_type)+' at '+str(offset))

    def __resolve_object(self, idx):
        try:
            return self.resolved[idx]
        except KeyError:
            obj = self.objects[idx]
            if isinstance(obj, list):
                new_array = []
                for i in obj:
                    new_array.append(self.__resolve_object(i))
                self.resolved[idx] = new_array
                return new_array
            if isinstance(obj, dict):
                new_dict = {}
                for key, value in obj.items():
                    resolved_key = self.__resolve_object(key)
                    resolved_value = self.__resolve_object(value)
                    new_dict[resolved_key] = resolved_value
                self.resolved[idx] = new_dict
                return new_dict
            else:
                self.resolved[idx] = obj
                return obj

    def parse(self):
        '''Parse the BPList'''
        # read header
        if self.data[:5] == b'<?xml':
            # load XML formated data and convert it
            plist_elements = {'bundles': {}}
            for bundle in plistlib.loads(self.data)['bundles']:
                plist_elements['bundles'].update({bundle['bundleName']: {'files': bundle['files']}})
            return plist_elements
        elif self.data[:8] != b'bplist00':
            raise Exception('Bad magic')

        # read trailer
        offset_size, object_ref_size, num_objects, top_obj, table_offset = struct.unpack(
            '!BB4xI4xI4xI', self.data[-26:]
        )
        self.offset_size = offset_size
        self.object_ref_size = object_ref_size
        self.number_of_objects = num_objects
        self.top_object = top_obj
        self.table_offset = table_offset
        self.top_object = self.table_offset - 8

        # read offset table
        self.offset_table = self.data[self.table_offset:-26]
        self.offsets = []
        offset_table = self.offset_table
        for i in range(self.number_of_objects):
            offset_entry = offset_table[:self.offset_size]
            offset_table = offset_table[self.offset_size:]
            self.offsets.append(self.__unpack_int_struct(self.offset_size, offset_entry))
        #print("** plist offsets: %s" % (self.offsets))

        # read object table
        self.objects = []
        k = 0
        for i in self.offsets:
            obj = self.__unpack_item(i)
            #print("** plist unpacked %d %s %s at %d" % (k, type(obj), obj, i))
            k += 1
            self.objects.append(obj)

        # rebuild object tree
        #for i in range(len(self.objects)):
        #    self.__resolveObject(i)

        # return root object
        return self.__resolve_object(self.number_of_objects-1)
